! Copyright 2019
!
! Alex G. Harvey with contributions from Danilo S. Brambila and Zdenek Masin.
!
! This file is part of UKRmol-out (UKRmol+ suite).
!
!     UKRmol-out is free software: you can redistribute it and/or modify
!     it under the terms of the GNU General Public License as published by
!     the Free Software Foundation, either version 3 of the License, or
!     (at your option) any later version.
!
!     UKRmol-out is distributed in the hope that it will be useful,
!     but WITHOUT ANY WARRANTY; without even the implied warranty of
!     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!     GNU General Public License for more details.
!
!     You should have received a copy of the GNU General Public License
!     along with  UKRmol-out (in source/COPYING). Alternatively, you can also visit
!     <https://www.gnu.org/licenses/>.
!
module dipelm_procs

   use dipelm_defs
   use dipelm_special_functions
   use dipelm_types
   use dipelm_smooth
   use dipelm_io
   use blas95_compak, only: gemm
   
   implicit none

contains

   subroutine  dipelm_drv(ifail)    
      implicit none

!     Arguments
      integer :: ifail 
                                                   
!     Local

!     Namelist variables
!     ------------------
!>    @defgroup DIPELMINP Namelist: &dipelminp
!>    @{
      integer              :: lu_pw_dipoles(maxprop_par)   = 0           !< File units holding partial wave dipoles from rsolve.
      integer              :: nset_pw_dipoles(maxprop_par) = 0           !< Set number of dipoles on each unit.
      character(len=11)    :: format_pw_dipoles            = 'FORMATTED' !< Partial wave dipoles file format.
      
      integer              :: output_style                 = 0           !< Choose output file style for randomly oriented molecule observables. 
                                                                         !< 0 =  All states in a single file.
                                                                         !< 1 =  A file for each (ion state, neutral state) pair.

      integer              :: iprint                       = 0           !< The higher the number the more is written to iwrite.
      integer              :: iwrite                       = 6           !< Destination for writing calculation details.   
      real(idp)            :: first_IP                     = 0._idp      !< Allows adjustment of 1st Ip. If zero calculated Ip is used.
      real(idp)            :: ion_charge                   = 1._idp      !< Charge of ionized molecule.
      integer              :: no_ion_states                = 0           !< No of ion states for which to calculate randomly oriented observables.
      integer              :: no_neutral_states            = 0           !< No of neutral states for which to calculate randomly oriented observables.
      integer, allocatable :: select_ion_states(:)                       !< Define subset of ion states for which to  calculate randomly oriented observables.
      integer, allocatable :: select_neutral_states(:)                   !< Define subset of neutrals states for which to  calculate randomly oriented observables.     
      logical              :: calc_oriented_observables    = .false.     !< Controls calculation of oriented obsesrvables. If .true. namelist orient nust be supplied.
      logical              :: smooth                       = .false.     !< Controls gaussian smoothing of dipoles. If .true. namelist smooth must be supplied.
      character(len=3)     :: point_group                  = ''          !< Not used currently.
!>    @}

!     For holding the dipoles      
      type(moments)     :: dipoles      
                             
!     Needed for reading DIPTRANS if we are using the old input style      
      integer   :: ngrdproj(2), ngrdalign(3), neutral_state
      real(idp) :: euler_angle_limits(6),scat_angle_limits(4)
      logical   :: legacy_input
      
      integer              :: ierr, i
      character(len=1000)  :: line
      
      integer              :: no_ion_states_all, no_neutral_states_all
      integer, allocatable :: tmp_arr1(:), tmp_arr2(:)
      logical              :: use_exp_IP
      real(idp)            :: calc_1st_IP 
            
!     For basis transformation      
      complex(idp), allocatable :: U(:,:), U1(:,:)
      integer                   :: lmax
      type(moments)             :: cmp_dipoles

!     Cross section and asymmetry parameters   
      real(idp), allocatable  :: xsec(:,:,:), beta_1(:,:,:), beta_2l(:,:,:), beta_2c(:,:,:)

      namelist /DIPELMINP/  lu_pw_dipoles,                        &
      &                     format_pw_dipoles,                    &
      &                     nset_pw_dipoles,                      &
      &                     output_style,                         &
      &                     iprint,                               &
      &                     iwrite,                               &
      &                     first_IP,                             &
      &                     ion_charge,                           &
      &                     no_ion_states,                        &
      &                     no_neutral_states,                    &
      &                     select_ion_states,                  &
      &                     select_neutral_states,              &          
      &                     calc_oriented_observables,            &   
      &                     smooth,                               &
      &                     point_group

!     Default values for DIPELMINP namelist
!     -------------------------------------

!      lu_pw_dipoles             = 0             
!      nset_pw_dipoles           = 0            
!      format_pw_dipoles         = 'FORMATTED'   ! 'UNFORMATTED' 
!      iprint                    = 0             
!      iwrite                    = 6             
!      first_IP                  = 0.0_idp       
!      ion_charge                = 1.0_idp      
!      no_ion_states             = 0            
!      no_neutral_states         = 0            
!      calc_oriented_observables = .false.                                     
!      smooth                    = .false.         
!      point_group               = ''          

      allocate(select_ion_states(max_states_for_dipoles), select_neutral_states(max_states_for_dipoles) )
      select_ion_states       = 0  
      select_neutral_states   = 0 

      legacy_input              = .false.        
!     Read namelists
!     --------------
      
!     Read dipelminp
            
      ierr = 0
      rewind(5)
      read(5, nml=DIPELMINP, iostat=ierr)
      
      if (ierr .ne. 0) then
         
         backspace(5)
         read(5,fmt='(A)') line
         write(iwrite,'(A)') &
           'Invalid line in namelist DIPELMINP: '//trim(line)
         write (iwrite,2000)
         ifail = 2000 
         
         return
           
      end if
    
!     For backwards compatability with old style of DIPELM input          

!     Check to see if DIPTRANS exists and if so get the needed variables
!     ( for random orientation these are first_IP, no_ion_states and neutral_state)

      call read_old_DIPTRANS( legacy_input, first_IP, no_ion_states, neutral_state, ngrdproj, ngrdalign, &
      &                       euler_angle_limits, scat_angle_limits )
      
      if (legacy_input) then         

!        Whether we smooth or not is controlled by the ismooth namelist 
!        variable in the old style input. The namelist smooth was always read.  
     
         smooth = .true. 
         
         write( iwrite,  '(/, " Old style of dipelm namelist input used. ")')   

      end if  

!     If first_IP is set, use it instead of calculated IP      
      use_exp_IP = .false. 
      if(first_IP .ne. 0.0_idp) use_exp_IP = .true. 
      
      
!     Read transition moments from file      
!     --------------------------------- 

      call read_moments( lu_pw_dipoles, format_pw_dipoles, nset_pw_dipoles, dipoles, iprint, &
      &                  iwrite, ifail )
      
      
      dipoles % mol % ion_charge = ion_charge ! Ion charge is not specified in the pw dipole files.
                                              ! TODO: This should be changed in the future
    
!     Determine how many and which states we want to calculate observables for

      no_ion_states_all      = size(dipoles % scattering_states % channels % ion_states % energies)
      no_neutral_states_all  = size(dipoles % neutral_states % energies)
      
      call move_alloc(select_ion_states, tmp_arr1)
      call move_alloc(select_neutral_states, tmp_arr2)
      
      allocate(select_ion_states(no_ion_states_all), select_neutral_states(no_neutral_states_all) )
      select_ion_states     = 0
      select_neutral_states = 0
      
      select_ion_states(1:no_ion_states)         = tmp_arr1(1:no_ion_states) 
      select_neutral_states(1:no_neutral_states) = tmp_arr2(1:no_neutral_states)
      
      call determine_states_selected( no_ion_states, select_ion_states, no_ion_states_all, &
   &                                  no_neutral_states, select_neutral_states, no_neutral_states_all)              
      
!     If we are using an experimental 1st IP then we adjust the neutral state energies accordingly
!     Energies of excited ion and neutral states relative to their respective ground states are 
!     not affected.      
     
      associate( E_neut => dipoles % neutral_states % energies, &
      &          E_ion  => dipoles % scattering_states % channels % ion_states % energies )
      
      calc_1st_IP = E_ion(1) - E_neut(1)
      
      if (use_exp_IP) then
         
         write( iwrite,  '(/, " Calculated Ip = ", D20.8)')  calc_1st_IP
         write( iwrite,  '(   " User input Ip = ", D20.8)')  first_IP           
 
         do  i = 1, size( E_neut )
         
            E_neut(i) = E_neut(i) + (calc_1st_IP - first_IP)
         
         end do
    
      else 

         write( iwrite,  '(/, " Calculated Ip = ", D20.8)')  calc_1st_IP 
                
      end if
      
      end associate

!     Smooth dipoles if requested
!     ---------------------------
            
      if (smooth) then
      
         call gaussian_smooth(dipoles, select_ion_states, select_neutral_states,  ifail)
      
      end if

!     -----------------------------------------------------      
!          
!     Calculate observables for randomly oriented molecules
!
!     -----------------------------------------------------      
      
!     Transform from basis of real spherical harmonics 
!     to complex spherical harmomnics      
!     ------------------------------------------------

      lmax = maxval( dipoles % scattering_states % channels % lchl )

      call sph_basis_transform_matrix(U, lmax, 'Slm')
      call sph_basis_transform_matrix(U1, 1, 'Slm')
      
!     Note that * is defined as matrix multiplication for the moments type, and a moment type is returned.
!>    @todo: Currently matmul is used for type(moment) B = type(moment) A * matrix, change to gemm.
      
      cmp_dipoles = conjg(U) * dipoles * transpose( U1(2:4,2:4) )

      cmp_dipoles % basis_type_electron = 'Ylm'
      cmp_dipoles % basis_type_photon   = 'Ylm'

!     Calculate Cross section and beta parameters.
!     --------------------------------------------

!>     @todo: Total xsection (summed over states) for unoriented molecules. 

      call calculate_cross_section_and_beta_parameters( cmp_dipoles, select_ion_states, select_neutral_states, &
      &                                                 xsec, beta_1, beta_2l, beta_2c )
      
      call write_cross_section_and_beta_parameters( dipoles, select_ion_states, select_neutral_states, output_style, &
      &                                             xsec, beta_1, beta_2l, beta_2c )
      
      write( iwrite,  '(/, " Unoriented cross sections and asymmetry parameters calculated. ")')      
      
!     Oriented observables.
!     ---------------------
      
      if (calc_oriented_observables) then

!        call oriented_observables(cmp_dipoles, legacy_input, iprint, iwrite, ifail) ! Ylm basis for electron and photon.
         call oriented_observables(dipoles, legacy_input, iprint, iwrite, ifail)     ! Slm basis for electron and photon.
      
      end if
      
      return
      
!     format statements
 2000 format(/,5X,'Problems reading the namelist DIPELMINP',/)    
   end subroutine dipelm_drv
   
   subroutine oriented_observables(dipoles, legacy_input, iprint, iwrite, ifail)   
      implicit  none
      
!     Arguments
      type(moments) :: dipoles
      logical       :: legacy_input
      integer       :: iprint, iwrite, ifail
      
!     Local

!     Namelist variables
!>    @defgroup ORIENT Namelist: &orient
!>    @{
      integer              :: ngrdproj(2)                    = [ 31, 4 ]    !< No. of photoelectron direction points (no_theta, no_phi)
      integer              :: ngrdalign(3)                   = [ 1, 1, 1 ]  !< No. of orientation (no_alpha, no_beta, no_gamma) points
      
                                                                            
      real(idp)            :: euler_angle_limits(6)          = [ 0._idp, 0._idp, 0._idp, 360._idp, 180._idp, 360._idp ]!<
                                                                            !< Defines the angular range for the molecular orientation
                                
      real(idp)            :: scat_angle_limits(4)           = [ 0._idp, 0._idp, 180._idp, 360._idp ] !<
                                                                            !< Defines the angular range for the photoelectron
                                                                            
      integer              :: no_ion_states_orient           = 0            !< No of ion states for which to calculate oriented observables
      integer              :: no_neutral_states_orient       = 0            !< No of neutral states for which to calculate oriented observables
      integer, allocatable :: select_ion_states_orient(:)                   !< Define subset of ion states for which to calculate oriented observables
      integer, allocatable :: select_neutral_states_orient(:)               !< Define subset of neutral states for which to calculate oriented observables
      logical              :: calc_mol_frame_observables     = .true.       !< Controls calculation of molecular frame observables (alpha = beta = gamma = 0)
      logical              :: calc_lab_frame_observables     = .false.      !< Controls calculation of lab frame observables
      logical              :: write_dipoles                  = .false.      !< Controls writing of Energy-angle form of the dipoles
!>    @}
      
      integer              :: lmax, no_channels, no_components, no_energies, no_ion_states_all, &
      &                       no_neutral_states_all, ien, ion ,ine, i, j, l, m, no_alpha, no_beta, &
      &                       no_gamma, no_euler, no_theta_phi, iangle, ia, ib, ig

      complex(idp), allocatable :: MFDip(:,:,:), LFDip(:,:,:,:), Ylm(:,:)
      real(idp),    allocatable :: MFPAD(:,:,:), LFPAD(:,:,:,:), grid_alpha(:), grid_beta(:), grid_gamma(:)
      
      integer   :: ierr
      
      real(idp) :: first_IP
      integer   :: neutral_state 
      
      namelist /ORIENT/   ngrdproj, &                    
     &                    ngrdalign,&                    
     &                    euler_angle_limits,&           
     &                    scat_angle_limits, &          
     &                    no_ion_states_orient, &               
     &                    no_neutral_states_orient, &           
     &                    select_ion_states_orient,&          
     &                    select_neutral_states_orient, &     
     &                    calc_mol_frame_observables, &  
     &                    calc_lab_frame_observables, &  
     &                    write_dipoles                  

!     Namelist ORIENT  defaults
!     --------------------------

!      ngrdproj                   = (/ 31, 4 /)
!      ngrdalign                  = (/ 1, 1, 1 /)
!      euler_angle_limits         = (/ 0._idp, 0._idp, 0._idp, 360._idp, 180._idp, 360._idp /)
!      scat_angle_limits          = (/ 0._idp, 0._idp, 180._idp, 360._idp /)
!      no_ion_states_orient              = 0
!      no_neutral_states_orient          = 0
!      calc_mol_frame_observables = .true. 
!      calc_lab_frame_observables = .false.
!      write_dipoles              = .false.
      
      no_ion_states_all      = size(dipoles % scattering_states % channels % ion_states % energies)
      no_neutral_states_all  = size(dipoles % neutral_states % energies)
      
      allocate(select_ion_states_orient(no_ion_states_all), select_neutral_states_orient(no_neutral_states_all) )
      select_ion_states_orient     = 0 
      select_neutral_states_orient = 0

      if(legacy_input) then
      
         call read_old_DIPTRANS( legacy_input, first_IP, no_ion_states_orient, neutral_state, ngrdproj, ngrdalign, &
         &                       euler_angle_limits, scat_angle_limits ) 
              
      else
      
         rewind(5)

         read(5,nml=ORIENT, iostat=ierr)
 
         if (ierr .gt. 0) then
            write (6,2001)
            ifail = 2001
            return
         end if
         
      end if
      
!     Process namelist input
!     ----------------------

!     Figure out how many and which states we should calculate oriented observables for.      

      call determine_states_selected( no_ion_states_orient, select_ion_states_orient, no_ion_states_all, &
   &                                  no_neutral_states_orient, select_neutral_states_orient, no_neutral_states_all)
      
!     Convert from degrees to radians

      euler_angle_limits = euler_angle_limits * 2*pi/360     
      scat_angle_limits  = scat_angle_limits  * 2*pi/360

!     Start calculation of oriented observables
!     -----------------------------------------      
      
!     Construct $i^{-l}e^(i\sigma_l) and multiply into dipoles.

      call multiply_in_coulomb_phase_factor(dipoles)

!     Calculate Molecular Frame Dipoles (MFDip) and Molecular Frame Photoelectron Angular 
!     Distributions (MFPAD)

      lmax  =  maxval( dipoles % scattering_states % channels % lchl )

      call grid_sp_harm(lmax, ngrdproj, scat_angle_limits, Ylm, dipoles % basis_type_electron)
    
      if (calc_mol_frame_observables) then

         do i = 1, no_neutral_states_orient
      
            ine = select_neutral_states_orient(i) 
      
            do j = 1, no_ion_states_orient
            
               ion = select_ion_states_orient(j)

               call calculate_oriented_dipoles( dipoles, ion, ine, 0._idp, 0._idp, 0._idp, Ylm, MFDip )
                
               if (write_dipoles) call write_MFDip( dipoles, scat_angle_limits, ngrdproj, ion, ine, MFDip ) 
      
               call calculate_MFPAD(dipoles, MFDip, ion, ine, MFPAD)
    
               call write_MFPAD( dipoles, scat_angle_limits, ngrdproj, ion, ine, MFPAD)
             
               deallocate(MFDip, MFPAD)
            
            end do
         
         end do
         
         if (write_dipoles) write( iwrite,  '(/, " MFDip calculated. ")') 
         write( iwrite,  '(/, " MFPAD calculated. ")')
                  
      end if
    
!     Calculate Lab Frame Dipoles (LFDip) and Lab Frame Photoelectron Angular 
!     Distributions (LFPAD). Warning output files quickly become large!
!     We loop over selected states, and orientations.      
      
      if (calc_lab_frame_observables) then

         no_theta_phi  = ngrdproj(1) * ngrdproj(2)
         no_alpha      = ngrdalign(1)
         no_beta       = ngrdalign(2)
         no_gamma      = ngrdalign(3)
         no_euler      = no_alpha*no_beta*no_gamma
         
         no_energies   = size( dipoles % scattering_states % energies )
         no_components = size( dipoles % scattering_states % moment_components )
          
!        First construct euler angle grids    
         call linspace( euler_angle_limits(1), euler_angle_limits(4), no_alpha, grid_alpha, include_end_point = .false. )
         call linspace( euler_angle_limits(2), euler_angle_limits(5), no_beta , grid_beta )
         call linspace( euler_angle_limits(3), euler_angle_limits(6), no_gamma, grid_gamma, include_end_point = .false. )            

         do i = 1, no_neutral_states_orient
      
            ine = select_neutral_states_orient(i) 
      
            do j = 1, no_ion_states_orient
            
               ion = select_ion_states_orient(j)
               
               allocate( LFDip(no_euler, no_theta_phi, no_components, no_energies), &
               &         LFPAD(no_euler, no_theta_phi, no_components, no_energies)  )
               iangle = 1
               
               do ig = 1, no_gamma
               do ib = 1, no_beta
               do ia = 1, no_alpha

                  call calculate_oriented_dipoles( dipoles, ion, ine, grid_alpha(ia), grid_beta(ib), grid_gamma(ig), Ylm, MFDip )
      
                  call calculate_MFPAD(dipoles, MFDip, ion, ine, MFPAD)
                  
                  LFDip(iangle,:,:,:) = MFDip
                  LFPAD(iangle,:,:,:) = MFPAD
                  
                  iangle = iangle + 1
            
                  deallocate(MFDip, MFPAD)
               
               end do
               end do
               end do
            
               if (write_dipoles) call write_LFDip( dipoles, scat_angle_limits, ngrdproj, & 
               &                                    grid_alpha, grid_beta, grid_gamma, ion, ine, LFDip )
               
               call write_LFPAD( dipoles, scat_angle_limits, ngrdproj, &
               &                 grid_alpha, grid_beta, grid_gamma, ion, ine, LFPAD )
               
               deallocate(LFDip, LFPAD)
            
            end do
         
         end do
         
         if (write_dipoles) write( iwrite,  '(/, " LFDip calculated. ")') 
         write( iwrite,  '(/, " LFPAD calculated. ")') 
              
      end if     

      return
!     format statements
 2001 format(/,5X,'Problems reading the namelist ORIENT',/)
   end subroutine
      
   subroutine calculate_oriented_dipoles( dipoles, ion_state, neutral_state, alpha, beta, gamma, Ylm, MFDip)
      implicit none
      
!     Arguments
      type(moments)             :: dipoles
      integer                   :: ion_state, neutral_state
      real(idp)                 :: alpha, beta, gamma      
      complex(idp)              :: Ylm(:,:)
      complex(idp), allocatable :: MFDip(:,:,:)      
!     Local

      
      integer                   :: lmax, no_channels, no_components, no_energies, no_ion_states, &
      &                            no_neutral_states, no_theta_phi
      integer                   :: i, it, ip, ien, ion, ine
      complex(idp), allocatable :: rot_Ylm(:,:), dk(:,:), R(:,:),  R1(:,:), U(:,:), Slm(:,:)
   
      associate ( D      => dipoles % elements )
      
      lmax               =  maxval( dipoles % scattering_states % channels % lchl )
      no_channels        =  size( D, 1 )
      no_components      =  size( D, 2 )
      no_energies        =  size( D, 3 )
      no_theta_phi       =  size(Ylm, 1)

      allocate( MFDip( no_theta_phi, no_components, no_energies) )      
      allocate( dk(no_theta_phi, no_components))

!     Create real or complex wigner D matrix elements depending on the basis type

      if (dipoles % basis_type_photon .eq. 'Ylm') then
      
         call a_wigner_d(1, alpha, beta, gamma, R1)
      
      else
      
         call a_re_wigner_d(1, alpha, beta, gamma, R1)
      
      end if
      
      if (dipoles % basis_type_electron .eq. 'Ylm') then
      
         call a_wigner_d(lmax, alpha, beta, gamma, R)
      
      else
      
         call a_re_wigner_d(lmax, alpha, beta, gamma, R)
      
      end if     

!     Rotate the spherical harmonics

!      allocate( rot_Ylm, mold = Ylm ) !F2008
      allocate( rot_Ylm( size(Ylm,1), size(Ylm,2) ) )
            
      call gemm( Ylm, R, rot_Ylm )          

!     Mulitply the rotated spherical harmonics into the dipoles and then 
!     rotate the dipole components 
         
      do ien = 1, no_energies
            
         call  gemm( rot_Ylm(:,:), D(:,:, ien, ion_state, neutral_state), dk )
               
         call  gemm( dk, R1(2:4,2:4), mfdip(:,:,ien), transb='T')

      end do
      
      end associate   
   end subroutine
      
   subroutine calculate_MFPAD(dipoles, MFDip, ion_state, neutral_state, MFPAD)
      implicit none
 
!     Argument variables
      type(moments)          :: dipoles
      complex(kind=idp)      :: MFDip(:,:,:)
      real(idp), allocatable :: MFPAD(:,:,:)
      
!     Local variables
      integer ::  no_ion_states, no_neutral_states, no_components, no_energies, no_angles, &
      &           ine, ion, ien, ico, ian, ion_state, neutral_state
      real(kind=idp) :: Aconst, photon_energy, Ip

      intent(in) :: dipoles
      
      no_angles            = size(MFDip,1)
      no_components        = size(MFDip,2)
      no_energies          = size(MFDip,3)
!      no_ion_states        = size(MFDip,4)
!      no_neutral_states    = size(MFDip,5)
        
      allocate( MFPAD(no_angles, no_components, no_energies) )
      
      associate( E_neut => dipoles % neutral_states % energies, &
      &          E_ion  => dipoles % scattering_states % channels % ion_states % energies, &
      &          E_elec => dipoles % scattering_states % energies )  
     
            
      Ip = E_ion(1) - E_neut( neutral_state ) 
    
      do  ien = 1, no_energies

         photon_energy = E_elec(ien) + Ip  
         Aconst        = 4*(pi**2)*alpha*photon_energy*convert_au_to_megabarns
               
         MFPAD(:,:,ien) = abs( MFDip(:,:,ien) )**2 * Aconst
               
      end do

      
      end associate
   
      return
! 1010 format(3(E20.5,x))
   end subroutine
   
   
   subroutine calculate_LFDip( dipoles, electron_direction, mol_orientation, steps_elec_direction, &
   &                           steps_mol_orientation, ion_state, neutral_state, MFDip)
      implicit none
!     Arguments
      type(moments)             :: dipoles
      integer                   :: steps_elec_direction(2),  steps_mol_orientation(3)
      real(idp)                 :: electron_direction(4),  mol_orientation(6)
      integer                   :: ion_state, neutral_state
      complex(idp), allocatable :: MFDip(:,:,:)      
!     Local

      
      integer                   :: lmax,no_channels, no_components, no_energies, no_ion_states, &
      &                            no_neutral_states, no_theta, no_phi, no_alpha, no_beta, no_gamma
      integer                   :: i, it, ip, ien, ion, ine, ia
      real(idp)                 :: alpha, beta, gamma
      complex(idp), allocatable :: Ylm(:,:), dk(:,:), R(:,:,:), R1(:,:,:), rot_Ylm(:,:)
   
      associate ( D      =>    dipoles % elements)
      
      lmax               =  maxval( dipoles % scattering_states % channels % lchl )
      no_channels        =  size( D, 1 )
      no_components      =  size( D, 2 )
      no_energies        =  size( D, 3 )
      
      no_theta   = steps_elec_direction(1)
      no_phi     = steps_elec_direction(2)
      no_alpha   = steps_mol_orientation(1)
      no_beta    = steps_mol_orientation(2)
      no_gamma   = steps_mol_orientation(3)

      allocate( MFDip( no_theta*no_phi, no_components, no_energies) )      
      allocate( dk(no_theta*no_phi, no_components))

      
      call grid_sp_harm(lmax, steps_elec_direction, electron_direction, Ylm, dipoles % basis_type_electron )
      
      call grid_wigner_D(lmax, steps_mol_orientation, mol_orientation, R, dipoles % basis_type_electron )
      
      call grid_wigner_D(1, steps_mol_orientation, mol_orientation, R1, dipoles % basis_type_photon )
      
!      allocate(rot_Ylm, mold = Ylm ) !F2008 standard
      allocate( rot_Ylm( size(Ylm,1), size(Ylm,1) ) )
            
      do ia = 1, no_alpha*no_beta*no_gamma

         call gemm(R(:,:,ia), Ylm, rot_Ylm )
         
         do ien = 1, no_energies
            
            call  gemm( rot_Ylm(:,:), D(:,:, ien, ion_state, neutral_state), dk )
               
            call  gemm( dk, R1(2:4,2:4, ia), mfdip(:,:,ien), transb='T')

         end do
         
      end do
      
      end associate   
   end subroutine  
       
   subroutine calculate_MFDip( dipoles, electron_direction, mol_orientation, steps_elec_direction, &
   &                           steps_mol_orientation, ion_state, neutral_state, MFDip)
      implicit none
!     Arguments
      type(moments)             :: dipoles
      integer                   :: steps_elec_direction(2),  steps_mol_orientation(3)
      real(idp)                 :: electron_direction(4),  mol_orientation(6)
      integer                   :: ion_state, neutral_state
      complex(idp), allocatable :: MFDip(:,:,:)      
!     Local

      
      integer                   :: lmax,no_channels, no_components, no_energies, no_ion_states, &
      &                            no_neutral_states, no_theta, no_phi, no_alpha, no_beta, no_gamma
      integer                   :: i, it, ip, ien, ion, ine
      real(idp)                 :: alpha, beta, gamma
      complex(idp), allocatable :: Ylm(:,:), dk(:,:),  R1(:,:), U(:,:), Slm(:,:)
   
      associate ( D      => dipoles % elements )
      
      lmax               =  maxval( dipoles % scattering_states % channels % lchl )
      no_channels        =  size( D, 1 )
      no_components      =  size( D, 2 )
      no_energies        =  size( D, 3 )
!      no_ion_states      =  size( D, 4 )
!      no_neutral_states  =  size( D, 5 )
      
      no_theta = steps_elec_direction(1)
      no_phi   = steps_elec_direction(2)

      allocate( MFDip( no_theta*no_phi, no_components, no_energies) )      
      allocate( dk(no_theta*no_phi, no_components))
      
      call grid_sp_harm(lmax, steps_elec_direction, electron_direction, Ylm, 'Ylm')
      
      alpha = mol_orientation(1)
      beta  = mol_orientation(2)
      gamma = mol_orientation(3)
      
      call a_wigner_d(1, alpha, beta, gamma, R1)     
      
      if (dipoles % basis_type_electron == 'Slm') then      
        
!        Would be more efficient to construct the real spherical harmonics directly
         call sph_basis_transform_matrix( U, lmax, 'Ylm' )
         allocate( Slm( no_theta*no_phi, no_channels ) )
         
         call gemm( Ylm, U, Slm, transb='T' )  
         Ylm = Slm   
         deallocate(Slm, U)
         
      end if
            
      if (dipoles % basis_type_photon == 'Slm') then
!        Would be more efficient to construct the real wigner d matrix elements directly      
         call sph_basis_transform_matrix(U, 1, 'Slm')
         
         R1 = matmul(R1, U)
         deallocate(U)
        
         call sph_basis_transform_matrix(U, 1, 'Ylm')
         R1 = matmul(U, R1)
         
      end if

         
      do ien = 1, no_energies
            
         call  gemm( Ylm(:,:), D(:,:, ien, ion_state, neutral_state), dk )
               
         call  gemm( dk, R1(2:4,2:4), mfdip(:,:,ien), transb='T')

      end do
      
      end associate   
   end subroutine
   
   subroutine calculate_cross_section_and_beta_parameters( dipoles, select_ion_states, select_neutral_states,&
   &                                                       xsec, beta_1, beta_2l, beta_2c)   
      implicit none
      
!     Modified by ZM to include calculation of the beta_1 parameter.
!     Arguments
      type(moments)          :: dipoles
      integer                :: select_ion_states(:), select_neutral_states(:)
      real(idp), allocatable :: xsec(:,:,:), beta_1(:,:,:), beta_2l(:,:,:), beta_2c(:,:,:)

!     Local variables 
      integer                   :: no_ion_states, no_neutral_states, no_energies, ion, ine, ien, &
      &                            Lbig,  no_ion_states_all, no_neutral_states_all, isne, ison     
      complex(idp), allocatable :: AL_zero(:), AL_one(:), AL_two(:), AL_two_c(:)
      real(idp)                 :: Ip, photon_energy, Aconst


      associate( E_neut => dipoles % neutral_states % energies, &
      &          E_ion  => dipoles % scattering_states % channels % ion_states % energies, &
      &          E_elec => dipoles % scattering_states % energies ) 

      no_ion_states_all     = size( E_ion  )
      no_neutral_states_all = size( E_neut )
      no_energies           = size( E_elec )
      
      no_ion_states         = size( select_ion_states )
      no_neutral_states     = size( select_neutral_states )
      
      allocate( AL_zero(no_energies), AL_one(no_energies), AL_two(no_energies), AL_two_c(no_energies) )
      
      allocate( xsec(no_energies, no_ion_states, no_neutral_states) )
      allocate( beta_1(no_energies, no_ion_states, no_neutral_states) )
      allocate( beta_2l(no_energies, no_ion_states, no_neutral_states) )
      allocate( beta_2c(no_energies, no_ion_states, no_neutral_states) )
      
      xsec   = 0._idp
      beta_1 = 0._idp
      beta_2l = 0._idp
      beta_2c = 0._idp
      
      
      do ine = 1, no_neutral_states_all
      
        if ( any(select_neutral_states .eq. ine) ) then
        
!         isne = findloc(select_neutral_states, ine, 1) !F2008 feature
         isne = maxloc(merge(1,0,select_neutral_states == ine),dim=1) !F2003 equivalent
      
         do ion = 1, no_ion_states_all
         
            if ( any(select_ion_states .eq. ion ) ) then
            
!            ison = findloc(select_ion_states, ion, 1) !F2008 feature
            ison = maxloc(merge(1,0,select_ion_states == ion), dim=1) !F2003 equivalent
         
            AL_zero        = 0._idp
            AL_one         = 0._idp           
            AL_two         = 0._idp
            AL_two_c       = 0._idp
            
            !>  @todo Check whether xsec and beta_2 depend on if the 
            !!       light polarization is circular or linear. (I think they actually might do)
            !!       Looks like beta_2 does. 

            !Linear polarization
            call make_ritchie_AL_coeff( dipoles, ion, ine, 0, 0, AL_zero )
            call make_ritchie_AL_coeff( dipoles, ion, ine, 2, 0, AL_two  )
            
            !Circular polarization            
            call make_ritchie_AL_coeff( dipoles, ion, ine, 1, 1, AL_one    )
            call make_ritchie_AL_coeff( dipoles, ion, ine, 2, 1, AL_two_c  )

            Ip = E_ion(1) - E_neut( ine )  

            do ien = 1, no_energies

               photon_energy = E_elec(ien) + Ip
               Aconst=(4._idp) * (pi**2) * alpha * photon_energy*convert_au_to_megabarns
         
               if (abs(AL_zero(ien)) .ne. 0) then
               
                  xsec(ien, ison, isne)    = real( AL_zero(ien), kind=idp ) * Aconst !Normalize to produce the integral cross section
                  
                  beta_1(ien, ison, isne)  = real( AL_one(ien)/AL_zero(ien), kind=idp )
               
                  beta_2l(ien, ison, isne) = real( AL_two(ien)/AL_zero(ien), kind=idp )
                  
                  beta_2c(ien, ison, isne) = real( AL_two_c(ien)/AL_zero(ien), kind=idp )
    
               end if
               
            end do
            
            end if
            
         end do
         
         end if
         
      end do

      end associate
   
   end subroutine

!
!  make_ritchie_AL_coeff
! 
!  B. Ritchie, PRA 13, 1411, (1976): formulae 11a,11b
!
!  On input p stands for the desired photon polarization p = -1,0,1 (left circ., linear, right circ.)
!  The calculated AL_coefficient does not include the normalization factor 
!  Aconst=(4._idp) * (pi**2) * alpha * photon_energy*convert_au_to_megabarns.
!  This can be applied outside this routine to ensure e.g. that the L=0 coefficient gives the 
!  correct integral cross section (see calculate_orientation_averaged_asymmetry_parameter). 
!     
   subroutine make_ritchie_AL_coeff( dipoles, ion_state, neutral_state, Lbig, p, AL_coefficient )
      implicit none
 
!     Arguments
      integer           :: ion_state, neutral_state, Lbig, p
      type(moments)     :: dipoles  
      complex(kind=idp) :: AL_coefficient(:)

!     Local
      integer                        :: no_energies, no_components, no_ion_states, &
      &                                 no_partials, no_neutral_states
      
      integer                        :: lmax, l, lp, ien, m, mp, ilpmp, ilm, q, qp
      
      real(kind=idp)                 :: k_final, charge, eta, pconst, cf_polarization, cf_angmom, &
      &                                 cf_space,cf_photon, coupling, Aconst, photon_energy, Ip,  &
      &                                 E_chan_thresh
      
      complex(kind=idp)              :: coeff_one, dipole_product
      
      complex(kind=icp), allocatable :: energy_cf(:)

      charge           = dipoles % mol % ion_charge
      AL_coefficient   = 0.0_idp
      
      associate ( D      => dipoles % elements, &
      &           E_elec => dipoles % scattering_states % energies, &
      &           E_ion  => dipoles % scattering_states % channels % ion_states % energies, &
      &           E_neut => dipoles % neutral_states % energies ) 

      no_components      =  size( dipoles % elements, 2 )
      no_energies        =  size( dipoles % elements, 3 )
      no_ion_states      =  size( dipoles % elements, 4 )
      no_neutral_states  =  size( dipoles % elements, 5 )
      
      allocate( energy_cf(no_energies) )
      
      lmax = maxval( dipoles % scattering_states % channels % lchl)

      cf_polarization = threej( 2*1, 2*p, 2*1, -2*p, 2*Lbig, 0) !set p=+-1 on input for circular polarization
      
      !Ip = E_ion(ion_state) - E_neut(neutral_state)
      
      do l = 0, lmax
      
         do lp = 0, lmax

            cf_angmom = threej( 2*l, 0, 2*lp, 0, 2*Lbig, 0 ) * (2*Lbig + 1.0_idp) &
            &         * sqrt( 2*l + 1.0_idp ) * sqrt( 2*lp + 1.0_idp )
            
            if ( cf_angmom .eq. 0.0_idp ) cycle

            energy_cf = 0.0_idp
            
            do ien = 1, no_energies
            
               E_chan_thresh = E_ion(ion_state) - E_ion(1) ! Electron energies are relative to the
                                                           ! ground state of the ion.
            
               if ( E_elec(ien) .ge.  E_chan_thresh ) then
               
                  k_final = sqrt( 2.0_idp * (E_elec(ien) - E_chan_thresh) )
               
                  if ( k_final .le. 0.0_idp ) exit
               
                  eta            = -charge/k_final                  
                  coeff_one      =  exp( eye*( CPHAZ(l,eta,6) - CPHAZ(lp,eta,6) ) ) * eye**(lp-l)  !< @todo Multiply coulomb phase into dipoles from the beginning                
                  !photon_energy  =  E_elec(ien)+Ip
                  !Aconst         =  1.0_idp !(4._idp) * (pi**2) * alpha * photon_energy*convert_au_to_megabarns
                  energy_cf(ien) =  coeff_one ! * Aconst
               
               end if
               
            end do

            do m = -l, l

               ilm = lm2i(l,m)

               do mp = -lp, lp

                  cf_space = threej( 2*l, 2*m, 2*lp, -2*mp, 2*Lbig, -2*(m-mp) )
                  
                  if (cf_space .eq. 0.0_idp) cycle

                  ilpmp = lm2i(lp,mp)

                  do q = -1, 1
                  
                     do qp = -1, 1

                        cf_photon = threej( 2*1, 2*q, 2*1, -2*qp, 2*Lbig, -2*(m-mp) )
                        
                        if (cf_photon .eq. 0.0_idp) cycle

                        coupling = (-1)**(p+q+m) * cf_angmom * cf_polarization * cf_space * cf_photon
                        
                        do ien = 1, no_energies
                        
                           dipole_product = D(ilm, 2+q, ien, ion_state, neutral_state) &
                                          * conjg( D(ilpmp, 2+qp, ien, ion_state, neutral_state) )
                                          
                           AL_coefficient(ien) = AL_coefficient(ien) + ( energy_cf(ien) & 
                           &                       * coupling * dipole_product )
                           
                        end do

                     end do
                     
                  end do
                   
               end do
               
            end do
            
         end do
         
      end do
      end associate
      
   end subroutine  make_ritchie_AL_coeff
   
   subroutine determine_states_selected( no_ion_states, select_ion_states, no_ion_states_all, &
   &                                     no_neutral_states, select_neutral_states, no_neutral_states_all)
      implicit none

!     Arguments      
      integer              :: no_ion_states, no_ion_states_all, no_neutral_states, no_neutral_states_all
      integer, allocatable :: select_ion_states(:), select_neutral_states(:)
!     Local      
      integer, allocatable :: tmp_arr(:)
      integer              :: i
   
!     Figure out how many and which states we should calculate oriented observables for.      
      if (no_ion_states .eq. 0) then
      
         no_ion_states = no_ion_states_all
        
         select_ion_states     = [ (i,i=1,no_ion_states,1) ]
      
      else if(no_ion_states .gt. 0) then

         if (select_ion_states(1) .eq. 0) then
             
             deallocate(select_ion_states)
             allocate( tmp_arr(no_ion_states) )
             tmp_arr = [ (i,i=1,no_ion_states,1) ]
          
             call move_alloc( tmp_arr, select_ion_states )

         else
         
             allocate( tmp_arr(no_ion_states) )     
             tmp_arr = select_ion_states(1:no_ion_states)             
             deallocate(select_ion_states)
             call move_alloc( tmp_arr, select_ion_states )
         
         end if
         
      end if

      if (no_neutral_states .eq. 0) then
      
         no_neutral_states = no_neutral_states_all
         select_neutral_states     = [ (i,i=1,no_neutral_states,1) ]
      
      else if(no_neutral_states .gt. 0) then
      
         if (select_neutral_states(1) .eq. 0) then
             
             deallocate(select_neutral_states)
             allocate( tmp_arr(no_neutral_states) )
             tmp_arr = [ (i,i=1,no_neutral_states,1) ]
             
             call move_alloc( tmp_arr, select_neutral_states )
             
         else
         
             allocate( tmp_arr(no_neutral_states) )         
             tmp_arr = select_neutral_states(1:no_neutral_states)
             deallocate(select_neutral_states)
             call move_alloc( tmp_arr, select_neutral_states )
         
         end if
         
      end if           
    
   end subroutine
   
!  Construct $i^{-l}e^(i\sigma_l) and multiply into dipoles.
!  ---------------------------------------------------------  
   subroutine multiply_in_coulomb_phase_factor(dipoles)
      implicit none
      
      type(moments) :: dipoles
      
      integer      :: lmax, no_channels, no_components, no_energies, no_ion_states, i, l, m, ien, ion
      real(idp)    :: E_chan_thresh, k_final, eta
      complex(idp) :: coulomb_phase

      associate ( D      => dipoles % elements, &
      &           E_elec => dipoles % scattering_states % energies, &
      &           E_ion  => dipoles % scattering_states % channels % ion_states % energies, &
      &           E_neut => dipoles % neutral_states % energies, &
      &           Z      => dipoles % mol % ion_charge ) 

      lmax               =  maxval( dipoles % scattering_states % channels % lchl )
      no_channels        =  size( D, 1 )
      no_components      =  size( D, 2 )
      no_energies        =  size( D, 3 )
      no_ion_states      =  size( D, 4 )


      do ion = 1, no_ion_states
         do ien = 1, no_energies
            
            E_chan_thresh = E_ion(ion) - E_ion(1)          ! Electron energies are relative to the
                                                           ! ground state of the ion.
            if ( E_elec(ien) .ge.  E_chan_thresh ) then
               
               k_final = sqrt( 2.0_idp * (E_elec(ien) - E_chan_thresh) )
               
               if ( k_final .le. 0.0_idp ) exit
               
               eta  = -Z/k_final  
               
               i=1
               do l = 0, lmax
               
                  coulomb_phase = (-eye)**l * exp( eye * CPHAZ(l,eta,6) )
                  
                  do m = -l, l
               
                     D(i,:,ien,ion,:) = D(i,:,ien,ion,:) * coulomb_phase
                     i = i + 1
                  
                  end do
               end do
            
            end if
            
         end do      
      end do
      
      end associate   
   
   end subroutine

end module
