! Copyright 2020
!
! For a comprehensive list of the developers that contributed to these codes
! see the UK-AMOR website.
!
! This file is part of UKRmol-out (UKRmol+ suite).
!
!     UKRmol-out is free software: you can redistribute it and/or modify
!     it under the terms of the GNU General Public License as published by
!     the Free Software Foundation, either version 3 of the License, or
!     (at your option) any later version.
!
!     UKRmol-out is distributed in the hope that it will be useful,
!     but WITHOUT ANY WARRANTY; without even the implied warranty of
!     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!     GNU General Public License for more details.
!
!     You should have received a copy of the GNU General Public License
!     along with  UKRmol-out (in source/COPYING). Alternatively, you can also visit
!     <https://www.gnu.org/licenses/>.
!
!> \brief   I/O routines used by MULTIDIP
!> \author  J Benda
!> \date    2020 - 2021
!>
!> This module contains routines that read the necessary input files (K-matrices, scattering coefficients, molecular_data files)
!> and return the needed subset of data.
!>
module multidip_io

    use, intrinsic :: iso_c_binding,   only: c_double, c_int, c_f_pointer
    use, intrinsic :: iso_fortran_env, only: input_unit, int32, int64, real64, output_unit, iostat_end

    ! GBTOlib
    use blas_lapack_gbl,  only: blasint
#ifdef WITH_MMAP
    use file_mapping_gbl, only: file_mapping
#endif
    use phys_const_gbl,   only: pi, imu, to_eV
    use precisn_gbl,      only: wp, wp_bytes

    implicit none

    integer :: myproc = 1
    integer :: nprocs = 1

    !> \brief   K-matrix file
    !> \author  J Benda
    !> \date    2020 - 2021
    !>
    !> This data structure contains data read from a K-matrix file produced by RSOLVE.
    !> K-matrices are used to obtain the final stationary photoionization wave function.
    !>
    type KMatrix
        integer  :: mgvn, stot, nescat, nchan, nopen, ndopen, nchsq, lukmt, nkset, maxne
        real(wp), allocatable :: escat(:)
    contains
        procedure :: get_kmatrix
        procedure :: read_kmatrix_file
        procedure :: reset_kmatrix_position
    end type KMatrix


    !> \brief   Photoionization wave function coefficients
    !> \author  J Benda
    !> \date    2020
    !>
    !> Photoionization Ak coefficients calculated and written by RSOLVE. This can be used
    !> as an alternative to calculating the Ak coefficients on the fly.
    !>
    type ScatAkCoeffs
        real(wp), allocatable :: ReA(:, :, :), ImA(:, :, :), evchl(:), escat(:)
        integer,  allocatable :: ichl(:), lvchl(:), mvchl(:)
        integer               :: mgvn, nesc, nchan, nstat, lusct
    contains
        procedure :: get_wfncoeffs
        procedure :: read_wfncoeffs_file
        procedure :: reset_wfncoeffs_position
    end type ScatAkCoeffs


    !> \brief  Auxiliary data structure for 3D matrix (potentially memory-mapped)
    !> \author J Benda
    !> \date   2021
    !>
    !> Matrix class that either contains allocated data, or pointer to a mapped memory.
    !> Used in MolecularData for inner dipole matrices and for boundary amplitude matrices.
    !>
    type MappedMatrix3D
        real(real64), pointer :: mat(:, :, :) => null()  !< pointer to mapped disk data
#ifdef WITH_MMAP
        type(file_mapping)    :: mapping                 !< helper structure for file memory mapping
#endif
    contains
        procedure :: load => load_mapped_matrix_3d
        final     :: destruct_mapped_matrix_3d
    end type MappedMatrix3D


    !> \brief   RMT molecular data file
    !> \author  J Benda
    !> \date    2020 - 2021
    !>
    !> This data structure contains data read from the molecular_data file produced by RMT_INTERFACE.
    !> Only a subset of values needed by this program is stored in memory. This in particular
    !> includes the inner and outer region transition dipole elements, boundary amplitudes and
    !> angular integrals of the real spherical harmonics.
    !>
    !> When the code is compiled with the WITH_MMAP=1 option, the dipole matrices will not be read
    !> into memory, but only mapped to the virtual memory.
    !>
    type MolecularData

        ! inner transition dipoles storage
        type(MappedMatrix3D)        :: dipx, dipy, dipz
        integer(int32), allocatable :: iidipx(:), iidipy(:), iidipz(:)
        integer(int32), allocatable :: ifdipx(:), ifdipy(:), ifdipz(:)

        ! boundary amplitudes
        type(MappedMatrix3D), allocatable :: wamp(:)

        ! other data
        real(real64)                :: rmatr
        real(real64),   allocatable :: crlv(:,:,:), eig(:,:), rg(:), etarg(:), gaunt(:, :, :)
        integer(int32), allocatable :: mgvns(:), mnp1(:), nchan(:), l2p(:,:), m2p(:,:), ichl(:,:), lm_rg(:,:)

    end type MolecularData

contains

    !> \brief  Read or map a 3D matrix from file
    !> \author J Benda
    !> \date   2021
    !>
    !> Depending on the compilation settings, either map the given section of the file to memory,
    !> or simply allocate and read the chunk.
    !>
    subroutine load_mapped_matrix_3d (this, filename, u, offset, a, b, c)

        class(MappedMatrix3D), intent(inout) :: this
        character(len=*),      intent(in)    :: filename
        integer(int32),        intent(in)    :: u, a, b, c
        integer(int64),        intent(in)    :: offset
        integer(int64)                       :: length

#ifdef WITH_MMAP
        length = int(a, int64) * int(b, int64) * int(c, int64) * wp_bytes
        call this % mapping % init(filename, offset - 1, length)
        call c_f_pointer(this % mapping % ptr, this % mat, [a, b, c])
#else
        allocate (this % mat(a, b, c))
        read (u) this % mat
#endif

    end subroutine load_mapped_matrix_3d


    !> \brief   Finalize a (potentially mapped) 3D matrix object
    !> \author  J Benda
    !> \date    2021
    !>
    subroutine destruct_mapped_matrix_3d (this)

        type(MappedMatrix3D), intent(inout) :: this

#ifndef WITH_MMAP
        if (associated(this % mat)) then
            deallocate (this % mat)
        end if
#endif
        nullify (this % mat)

    end subroutine destruct_mapped_matrix_3d


    !> \brief   Read K-matrix files
    !> \author  J Benda
    !> \date    2020 - 2021
    !>
    !> Read all needed K-matrix files. These are needed to calculate scattering coefficients (Ak)
    !> during the calculation.
    !>
    subroutine read_kmatrices (km, lukmt, nkset)

        type(KMatrix), allocatable, intent(inout) :: km(:)
        integer,                    intent(in)    :: lukmt(:), nkset(:)

        integer :: i

        do i = 1, size(km)
            km(i) % lukmt = lukmt(i)
            km(i) % nkset = nkset(i)
            call km(i) % read_kmatrix_file
        end do

    end subroutine read_kmatrices


    !> \brief   Read K-matrix file
    !> \author  J Benda
    !> \date    2021
    !>
    !> Read metadata from a K-matrix file, count K-matrices and store the unit and dataset for
    !> later use in retrieval of the K-matrices themselves. The K-matrices are *not* being read
    !> into memory here to avoid exhausting RAM (particularly in parallel mode) if the K-matrices
    !> are large.
    !>
    subroutine read_kmatrix_file (km)

        class(KMatrix), intent(inout) :: km

        integer :: k, n, ifail, key, nset, nrec, ninfo, ndata, ntarg, nvib, ndis, ie, gutot, ion
        integer :: nopen, ndopen, nchsq, a, b, c, mye, nene, ierr
        logical :: iwarn
        real(wp) :: dE1, dE2, r, rmass, en
        real(wp), allocatable :: kmat(:)
        character(len=80) :: title

        print '(/,A,I0,A,I0,/)', 'Reading K-matrices from unit ', km % lukmt, ', set ', km % nkset

        call getset(km % lukmt, km % nkset, 11, 'UNFORMATTED', ifail)

        read (km % lukmt) key, nset, nrec, ninfo, ndata
        read (km % lukmt) title
        read (km % lukmt) km % mgvn, km % stot, gutot, ion, r, rmass
        read (km % lukmt) ntarg, nvib, ndis, km % nchan, km % maxne

        print '(2x,A,I0)', 'mgvn  = ', km % mgvn
        print '(2x,A,I0)', 'stot  = ', km % stot
        print '(2x,A,I0)', 'ntarg = ', ntarg
        print '(2x,A,I0)', 'nvib  = ', nvib
        print '(2x,A,I0)', 'ndis  = ', ndis
        print '(2x,A,I0)', 'nchan = ', km % nchan
        print '(2x,A,I0)', 'maxne = ', km % maxne

        print '(/,2x,A,/)', 'Energy ranges:'

        km % nescat = 0
        do ie = 1, km % maxne
            read (km % lukmt) k, n, dE1, dE2
            print '(4x,I4,I8,F8.3,F8.3)', k, n, dE1, dE2
            km % nescat = km % nescat + n
        end do

        nene = (km % nescat + nprocs - 1) / nprocs
        allocate (km % escat(km % nescat), kmat(km % nchan * (km % nchan + 1)))

        iwarn = .true.
        mye = 0
        do ie = 1, km % nescat
            read (km % lukmt, iostat = ierr) nopen, ndopen, nchsq, en!, kmat(1:nchsq)
            if (ierr == iostat_end) then
                print '(/,2x,A,I0,A,I0,A)', 'WARNING: Unit ', km % lukmt, ' contains only ', ie - 1, ' K-matrices.'
                print '(11x,A)', 'This means that CFASYM failed during RSOLVE for some energies.'
                print '(11x,A)', 'The scattering energies MUST be consistent across symmetries!'
                km % nescat = ie - 1
                exit
            else if (ierr /= 0) then
                print '(/,2x,A,I0,A)', 'ERROR ', ierr, ' while reading K-matrix file.'
                stop 1
            end if
            km % escat(ie) = en/2  ! Ry -> a.u.
            if (iwarn .and. nopen < km % nchan) then
                print '(/,2x,A,I0,A)', 'WARNING: Unit ', km % lukmt, ' contains only open-open subset of the full K-matrix.'
                print '(11x,A)', 'Use IKTYPE = 1 in RSOLVE to save full K-matrices.'
                print '(11x,A)', 'You may get spurious below-threshold noise.'
                iwarn = .false.
            end if
        end do

        deallocate (kmat)

    end subroutine read_kmatrix_file


    !> \brief   Reset I/O pointer to start of K-matrices
    !> \author  J Benda
    !> \date    2021
    !>
    !> Rewind the unit to the start of the associated dataset and read through to the very beginning of the actual
    !> K-matrix data. Also skip the given number of leading K-matrices. This prepares the file for the subsequent
    !> calls to `get_kmatrix`.
    !>
    subroutine reset_kmatrix_position (km, skip)

        class(KMatrix),    intent(in) :: km
        integer, optional, intent(in) :: skip

        integer :: i

        ! rewind to the beginning of the K-matrix unit
        call getset(km % lukmt, km % nkset, 11, 'UNFORMATTED', i)

        read (km % lukmt)
        read (km % lukmt)
        read (km % lukmt)
        read (km % lukmt)

        do i = 1, km % maxne
            read (km % lukmt)
        end do

        if (present(skip)) then
            do i = 1, skip
                read (km % lukmt)
            end do
        end if

    end subroutine reset_kmatrix_position


    !> \brief   Read single K-matrix from the K-matrix file
    !> \author  J Benda
    !> \date    2021
    !>
    !> Assuming that the K-matrix file associated with this object is correctly positioned, read the next K-matrix record.
    !> Symmetrize the K-matrix and store it into the allocatable two-dimensional array `kmat` (re/allocate as necessary).
    !>
    subroutine get_kmatrix (km, kmat, skip)

        class(KMatrix),        intent(in)    :: km
        real(wp), allocatable, intent(inout) :: kmat(:, :)
        integer,  optional,    intent(in)    :: skip

        real(wp), allocatable :: buffer(:)

        real(wp) :: en
        integer  :: a, b, c, i, ierr, nopen, ndopen, nchsq

        allocate (buffer(km % nchan * (km % nchan + 1)))

        read (km % lukmt, iostat = ierr) nopen, ndopen, nchsq, en, buffer(1:nchsq)

        if (ierr == iostat_end) then
            print '(/,2x,A,I0,A)', 'ERROR: Unit ', km % lukmt, ' ended while reading K-matrix!'
            stop 1
        else if (ierr /= 0) then
            print '(/,2x,A,2(I0,A))', 'ERROR ', ierr, ' while reading K-matrix file unit ', km % lukmt, '!'
            stop 1
        end if

        if (allocated(kmat)) then
            if (size(kmat, 1) /= km % nchan .or. size(kmat, 2) /= km % nchan) then
                deallocate (kmat)
            end if
        end if

        if (.not. allocated(kmat)) then
            allocate (kmat(km % nchan, km % nchan))
        end if

        c = 0
        do a = 1, nopen
            do b = 1, a
                c = c + 1
                kmat(a, b) = buffer(c)
                kmat(b, a) = buffer(c)
            end do
        end do

        if (present(skip)) then
            do i = 1, skip
                read (km % lukmt, iostat = ierr)  ! ignore EOF
            end do
        end if

    end subroutine get_kmatrix


    !> \brief   Read wave function coefficients from files
    !> \author  J Benda
    !> \date    2020
    !>
    !> Read wave function (Ak) coefficients from an unformatted file written by RSOLVE. This is an optional
    !> functionality meant for dabugging. It is recommended that the Ak coefficients are calculated by the
    !> present program instead to avoid storing large datasets on disk or in memory.
    !>
    subroutine read_wfncoeffs (ak, lusct)

        type(ScatAkCoeffs), allocatable, intent(inout) :: ak(:)
        integer, intent(in) :: lusct(:)
        integer :: i

        do i = 1, size(ak)
            ak(i) % lusct = lusct(i)
            call ak(i) % read_wfncoeffs_file
        end do

    end subroutine read_wfncoeffs


    !> \brief   Read wave function coefficients for a single symmetry from a file
    !> \author  J Benda
    !> \date    2021
    !>
    subroutine read_wfncoeffs_file (ak)

        class(ScatAkCoeffs), intent(inout) :: ak

        real(wp), allocatable :: ReA(:), ImA(:)
        real(wp)              :: rr
        integer               :: i, j, keysc, nset, nrec, ninfo, ndata, stot, gutot, nscat, ich, ie, mye, nene
        character(len=80)     :: header

        print '(/,A,I0,/)', 'Reading wave function coefficients from unit ', ak % lusct

        open (ak % lusct, status = 'old', action = 'read', form = 'unformatted')

        read (ak % lusct) keysc, nset, nrec, ninfo, ndata
        read (ak % lusct) header
        read (ak % lusct) nscat, ak % mgvn, stot, gutot, ak % nstat, ak % nchan, ak % nesc

        print '(2x,A,I0)', 'mgvn  = ', ak % mgvn
        print '(2x,A,I0)', 'stot  = ', stot
        print '(2x,A,I0)', 'nstat = ', ak % nstat
        print '(2x,A,I0)', 'nchan = ', ak % nchan
        print '(2x,A,I0)', 'nesc  = ', ak % nesc

        read (ak % lusct) rr

        allocate (ak % ichl(ak % nchan), &
                  ak % lvchl(ak % nchan), &
                  ak % mvchl(ak % nchan), &
                  ak % evchl(ak % nchan))

        read (ak % lusct) ak % ichl, ak % lvchl, ak % mvchl, ak % evchl

        print '(/,2x,A)', 'channel table'
        do j = 1, ak % nchan
            print '(2x,4I6,F15.7)', j, ak % ichl(j), ak % lvchl(j), ak % mvchl(j), ak % evchl(j)
        end do

        close (ak % lusct)

    end subroutine read_wfncoeffs_file


    !> \brief   Reset I/O pointer to start of Ak-coeffs
    !> \author  J Benda
    !> \date    2021
    !>
    !> Rewind the unit to the start of the associated dataset and read through to the very beginning of the actual
    !> Ak-coeffs data. Also skip the given number of leading Ak-coeffs. This prepares the file for the subsequent
    !> calls to `get_wfncoeffs`.
    !>
    subroutine reset_wfncoeffs_position (ak, skip)

        class(ScatAkCoeffs), intent(in) :: ak
        integer, optional,   intent(in) :: skip

        integer :: i, ich

        ! rewind to the beginning of the K-matrix unit
        call getset(ak % lusct, 1, 88, 'UNFORMATTED', i)

        read (ak % lusct)
        read (ak % lusct)
        read (ak % lusct)
        read (ak % lusct)
        read (ak % lusct)

        if (present(skip)) then
            do i = 1, skip
                do ich = 1, ak % nchan
                    read (ak % lusct)
                    read (ak % lusct)
                end do
            end do
        end if

    end subroutine reset_wfncoeffs_position


    !> \brief   Read single Ak-matrix from the Ak-coeffs file
    !> \author  J Benda
    !> \date    2021
    !>
    !> Assuming that the Ak-coeffs file associated with this object is correctly positioned, read the next Ak-matrix record.
    !>
    subroutine get_wfncoeffs (ak, Re_Ak, Im_Ak, skip)

        class(ScatAkCoeffs),   intent(in)    :: ak
        real(wp), allocatable, intent(inout) :: Re_Ak(:, :), Im_Ak(:, :)
        integer,  optional,    intent(in)    :: skip

        real(wp) :: E
        integer  :: i, ich, ierr

        do ich = 1, ak % nchan
            read (ak % lusct, iostat = ierr) E, i, Re_Ak(:, ich)
            read (ak % lusct, iostat = ierr) E, i, Im_Ak(:, ich)
            if (ierr == iostat_end) then
                print '(/,2x,A,I0,A)', 'ERROR: Unit ', ak % lusct, ' ended while reading Ak-coeffs!'
                stop 1
            else if (ierr /= 0) then
                print '(/,2x,A,2(I0,A))', 'ERROR ', ierr, ' while reading Ak-coeffs file unit ', ak % lusct, '!'
                stop 1
            end if
        end do

        if (present(skip)) then
            do i = 1, skip
                do ich = 1, ak % nchan
                    read (ak % lusct, iostat = ierr)  ! ignore EOF
                    read (ak % lusct, iostat = ierr)  ! ignore EOF
                end do
            end do
        end if

    end subroutine get_wfncoeffs


    !> \brief   Read RMT molecular_data file
    !> \author  J Benda
    !> \date    2020
    !>
    !> Read data from the RMT molecular_data file. This includes in particular both the inner and outer region
    !> transition dipole elements, as well as for instance Gaunt angular integrals for all partial waves used.
    !>
    subroutine read_molecular_data (moldat, filename)

        use multidip_params, only: check_dipoles

        type(MolecularData), intent(inout) :: moldat
        character(*),        intent(in)    :: filename

        integer(int32), allocatable :: ltarg(:), starg(:), nconat(:)
        real(real64),   allocatable :: r_points(:), cf(:, :, :)

        integer(int32) :: i, j, u, s, s1, s2, ntarg, nrg, nelc, nz, lrang2, lamax, inast, nchmx, nstmx, lmaxp1, nfdm
        integer(int32) :: lrgl, nspn, npty
        integer(int64) :: offset, length
        real(real64)   :: bbloch, delta_r

        open (newunit = u, file = filename, access = 'stream', status = 'old', action = 'read')

        print '(/,3A,/)', 'Reading file "', filename, '"'

        read (u) s, s1, s2
        length = s * int(s1, int64) * int(s2, int64) * wp_bytes
        allocate (moldat % iidipy(s), moldat % ifdipy(s))
        read (u) moldat % iidipy, moldat % ifdipy
        inquire (u, pos = offset)
        call moldat % dipy % load(filename, u, offset, s1, s2, s)
        offset = offset + length
        print '(2x,A,*(1x,I0))', '[m = -1] iidip =', moldat % iidipy
        print '(2x,A,*(1x,I0))', '         ifdip =', moldat % ifdipy

        if (check_dipoles) then
            if (sum(abs(moldat % dipy % mat)) == 0) then
                print '(/,2x,A,/)', 'WARNING: Norm of all Y dipoles is zero. Something went wrong in CDENPROP?'
            end if
        end if

        read (u, pos = offset) s, s1, s2
        length = s * int(s1, int64) * int(s2, int64) * wp_bytes
        allocate (moldat % iidipz(s), moldat % ifdipz(s))
        read (u) moldat % iidipz, moldat % ifdipz
        inquire (u, pos = offset)
        call moldat % dipz % load(filename, u, offset, s1, s2, s)
        offset = offset + length
        print '(2x,A,*(1x,I0))', '[m =  0] iidip =', moldat % iidipz
        print '(2x,A,*(1x,I0))', '         ifdip =', moldat % ifdipz

        if (check_dipoles) then
            if (sum(abs(moldat % dipz % mat)) == 0) then
                print '(/,2x,A,/)', 'WARNING: Norm of all Z dipoles is zero. Something went wrong in CDENPROP?'
            end if
        end if

        read (u, pos = offset) s, s1, s2
        length = s * int(s1, int64) * int(s2, int64) * wp_bytes
        allocate (moldat % iidipx(s), moldat % ifdipx(s))
        read (u) moldat % iidipx, moldat % ifdipx
        inquire (u, pos = offset)
        call moldat % dipx % load(filename, u, offset, s1, s2, s)
        offset = offset + length
        print '(2x,A,*(1x,I0))', '[m = +1] iidip =', moldat % iidipx
        print '(2x,A,*(1x,I0))', '         ifdip =', moldat % ifdipx

        if (check_dipoles) then
            if (sum(abs(moldat % dipx % mat)) == 0) then
                print '(/,2x,A,/)', 'WARNING: Norm of all X dipoles is zero. Something went wrong in CDENPROP?'
            end if
        end if

        read (u, pos = offset) ntarg
        print '(/,2x,A,I0)', 'ntarg = ', ntarg
        allocate (moldat % crlv(ntarg, ntarg, 3))
        read (u) moldat % crlv

        read (u) nrg
        allocate (moldat % rg(nrg), moldat % lm_rg(6, nrg))
        read (u) moldat % rg, moldat % lm_rg

        read (u) nelc, nz, lrang2, lamax, ntarg, inast, nchmx, nstmx, lmaxp1
        read (u) moldat % rmatr, bbloch
        allocate (moldat % etarg(ntarg), ltarg(ntarg), starg(ntarg))
        read (u) moldat % etarg, ltarg, starg
        read (u) nfdm, delta_r
        allocate (r_points(nfdm + 1))
        read (u) r_points
        print '(2x,A,*(1x,F8.3))', 'sampling radii:', r_points

        print '(/,2x,A)', 'number of inner eigenstates per symmetry'

        allocate (nconat(ntarg), moldat % l2p(nchmx, inast), moldat % m2p(nchmx, inast), moldat % eig(nstmx, inast), &
                  moldat % wamp(inast), cf(nchmx, nchmx, lamax), moldat % ichl(nchmx, inast), &
                  moldat % mgvns(inast), moldat % nchan(inast), moldat % mnp1(inast))

        inquire (u, pos = offset)

        length = int(nchmx, int64) * int(nstmx, int64) * wp_bytes

        do i = 1, inast

            read (u, pos = offset) lrgl, nspn, npty, moldat % nchan(i), moldat % mnp1(i)
            moldat % mgvns(i) = lrgl - 1
            print '(4x,I3,I8)', moldat % mgvns(i), moldat % mnp1(i)
            read (u) nconat, moldat % l2p(:, i), moldat % m2p(:, i)
            read (u) moldat % eig(:, i)
            inquire (u, pos = offset)
            call moldat % wamp(i) % load(filename, u, offset, nchmx, nstmx, 1_int32)
            offset = offset + length
            read (u, pos = offset) cf, moldat % ichl(:, i)
            read (u) s1, s2

            ! skip wmat2 (not used in MULTIDIP)
            inquire (u, pos = offset)
            offset = offset + int(s1, int64) * int(s2, int64) * nfdm * wp_bytes

        end do

        do i = 1, inast
            print '(/,2x,A,I0,A,I0,A)', 'channels in symmetry #', i, ' (IRR ', moldat % mgvns(i), '):'
            do j = 1, moldat % nchan(i)
                print '(2x,I4,A,I6,I3,SP,I6)', j, ':', moldat % ichl(j, i), moldat % l2p(j, i), moldat % m2p(j, i)
            end do
        end do

        close (u)

        call setup_angular_integrals(moldat)

    end subroutine read_molecular_data


    !> \brief   Store the angular integrals in a more convenient shape
    !> \author  J Benda
    !> \date    2020
    !>
    !> Convert the Gaunt integral storage from a list to an easily addressable matrix.
    !>
    subroutine setup_angular_integrals (moldat)

        type(MolecularData), intent(inout) :: moldat

        integer :: maxl, i, l1, m1, l2, m2, l3, m3

        maxl = maxval(moldat % lm_rg)

        allocate (moldat % gaunt((maxl + 1)**2, (maxl + 1)**2, -1:1))

        moldat % gaunt = 0

        do i = 1, size(moldat % rg)
            l1 = moldat % lm_rg(1, i); m1 = moldat % lm_rg(2, i)
            l2 = moldat % lm_rg(3, i); m2 = moldat % lm_rg(4, i)
            l3 = moldat % lm_rg(5, i); m3 = moldat % lm_rg(6, i)
            if (l3 /= 1 .or. abs(m3) > 1) then
                print '(A)', 'Unexpected layout of the angular integral storage in molecular_data'
                stop
            end if
            moldat % gaunt(l1*(l1+1) + m1 + 1, l2*(l2+1) + m2 + 1, m3) = moldat % rg(i)
        end do

    end subroutine setup_angular_integrals


    !> \brief   Return dipole transition descriptors
    !> \author  J Benda
    !> \date    2020
    !>
    !> Based on the Cartesian component index I = 1, 2, 3, (re)allocate the arrays 'iidip' and 'ifdip'
    !> and fill them with initial/final irreducible representations pairs for dipole matrices present
    !> in the molecular_data file structure 'moldat'.
    !>
    subroutine get_diptrans (moldat, I, iidip, ifdip)

        type(MolecularData),  intent(in)    :: moldat
        integer,              intent(in)    :: I
        integer, allocatable, intent(inout) :: iidip(:), ifdip(:)

        if (allocated(iidip)) deallocate (iidip)
        if (allocated(ifdip)) deallocate (ifdip)

        select case (I)
            case (1)
                iidip = moldat % iidipx
                ifdip = moldat % ifdipx
            case (2)
                iidip = moldat % iidipy
                ifdip = moldat % ifdipy
            case (3)
                iidip = moldat % iidipz
                ifdip = moldat % ifdipz
        end select

    end subroutine get_diptrans


    !> \brief   Multiply by dipole matrix
    !> \author  J Benda
    !> \date    2020
    !>
    !> Multiply matrix 'X' by the dipole matrix corresponding to dipole component 'I' and transition index 's'.
    !> Store the result in matrix 'Y'. Use 'nn' rows of 'X', and 'nf' rows of 'Y'.
    !>
    subroutine apply_dipole_matrix (moldat, I, s, transp, nf, nn, X, Y)

        use multidip_params,  only: rone, rzero
        use multidip_special, only: blas_dgemm => dgemm

        type(MolecularData), intent(in)    :: moldat
        integer,             intent(in)    :: I, s, nf, nn
        character(len=1),    intent(in)    :: transp
        real(wp),            intent(inout) :: X(:, :, :), Y(:, :, :)

        integer(blasint)  :: lda, ldb, ldc, m, n, k
        real(wp), pointer :: dips(:, :)

        ldb = size(X, 1)
        ldc = size(Y, 1)

        m = nf
        n = size(X, 2) * size(X, 3)
        k = nn

        select case (I)
            case (1)
                lda = size(moldat % dipx % mat, 1)
                dips => moldat % dipx % mat(:, :, s)
                call blas_dgemm(transp, 'N', m, n, k, rone, dips, lda, X, ldb, rzero, Y, ldc)
            case (2)
                lda = size(moldat % dipy % mat, 1)
                dips => moldat % dipy % mat(:, :, s)
                call blas_dgemm(transp, 'N', m, n, k, rone, dips, lda, X, ldb, rzero, Y, ldc)
            case (3)
                lda = size(moldat % dipz % mat, 1)
                dips => moldat % dipz % mat(:, :, s)
                call blas_dgemm(transp, 'N', m, n, k, rone, dips, lda, X, ldb, rzero, Y, ldc)
        end select

    end subroutine apply_dipole_matrix


    !> \brief   Multiply by boundary amplitudes
    !> \author  J Benda
    !> \date    2020
    !>
    !> Multiply real matrix X by the matrix of boundary amplitudes and store the result into the real
    !> matrix Y. When transp is 'T', then the elements of the matrix X are expected to correspond to outer
    !> region channels, while the elements of the matrix Y to the inner region states. Otherwise,
    !> the opposite is assumed.
    !>
    subroutine apply_boundary_amplitudes (moldat, irr, transp, X, Y)

        use multidip_special, only: blas_dgemm => dgemm

        type(MolecularData),   intent(in)    :: moldat
        integer,               intent(in)    :: irr
        character(len=1),      intent(in)    :: transp
        real(wp),              intent(inout) :: X(:,:), Y(:,:)

        integer(blasint)  :: m, n, k, lda, ldb, ldc
        real(wp)          :: alpha = 1, beta = 0
        real(wp), pointer :: wamp(:, :)

        if (transp == 'T') then
            m = moldat % mnp1(irr)
            n = size(X, 2)
            k = moldat % nchan(irr)
        else
            m = moldat % nchan(irr)
            n = size(X, 2)
            k = moldat % mnp1(irr)
        end if

        lda = size(moldat % wamp(irr) % mat, 1)
        ldb = size(X, 1)
        ldc = size(Y, 1)

        wamp => moldat % wamp(irr) % mat(:, :, 1)

        call blas_dgemm(transp, 'N', m, n, k, alpha, wamp, lda, X, ldb, beta, Y, ldc)

    end subroutine apply_boundary_amplitudes


    !> \brief   Scale boundary amplitudes matrix by a diagonal matrix
    !> \author  J Benda
    !> \date    2020
    !>
    !> Multiply boundary amplitudes for each inner region state by an element of the provided vector 'v'.
    !> Writes the result into 'vw'.
    !>
    subroutine scale_boundary_amplitudes (moldat, irr, v, vw)

        type(MolecularData),   intent(in)    :: moldat
        integer,               intent(in)    :: irr
        real(wp), allocatable, intent(in)    :: v(:)
        real(wp), allocatable, intent(inout) :: vw(:, :)

        integer :: k, nstat, p, nchan

        nstat = moldat % mnp1(irr)
        nchan = moldat % nchan(irr)

        do concurrent (p = 1:nchan, k = 1:nstat)
            vw(k, p) = v(k) * moldat % wamp(irr) % mat(p, k, 1)
        end do

    end subroutine scale_boundary_amplitudes


    !> \brief   Write partial wave moments
    !> \author  J Benda
    !> \date    2020 - 2021
    !>
    !> Produce tables with the multi-photon ionisation matrix elements per partial wave.
    !>
    subroutine write_partial_wave_moments (moldat, M, nesc, suffix)

        type(MolecularData),      intent(in) :: moldat
        character(len=*),         intent(in) :: suffix
        integer,                  intent(in) :: nesc
        complex(wp), allocatable, intent(in) :: M(:, :, :)

#ifdef WITH_COARRAYS
        complex(wp), allocatable :: val(:)[:]
#else
        complex(wp), allocatable :: val(:)
#endif

        integer            :: ie, irr, mgvn, u, ichan, ichlf, lf, mf, mye
        character(len=100) :: filename

#ifdef WITH_COARRAYS
        allocate (val(nesc)[*])
#else
        allocate (val(nesc))
#endif

        ! there is one file per irreducible representation and channel
        do irr = 1, size(moldat % mgvns)
            mgvn = moldat % mgvns(irr)
            do ichan = 1, size(M, 1)
                val = 0
                mye = 0
                do ie = myproc, nesc, nprocs
                    mye = mye + 1
                    val(ie) = M(ichan, mye, mgvn)
                end do
#ifdef WITH_COARRAYS
                sync all
                call co_sum(val, 1)  ! gather coarray from all images to the first one
                sync all
#endif
                ! the first image performs the writing
                if (myproc == 1 .and. any(val /= 0)) then
                    ichlf = moldat % ichl(ichan, irr)
                    lf = moldat % l2p(ichan, irr)
                    mf = moldat % m2p(ichan, irr)
                    write (filename, '(3(A,I0),A,SP,I0,3A)') 'pwdip-', mgvn, '-(', ichlf, ',', lf, ',', mf, ')', suffix, '.txt'
                    open (newunit = u, file = filename, action = 'write', form = 'formatted')
                    write (u, '(2E25.15)') val
                    close (u)
                end if
            end do
        end do

    end subroutine write_partial_wave_moments


    !> \brief   Write raw transition dipole moments
    !> \author  J Benda
    !> \date    2021
    !>
    !> Produce tables with the multi-photon ionisation matrix elements per partial wave and per ionisation chain.
    !>
    subroutine write_raw_dipoles (M, chains, nesc, stem)

        character(len=*),         intent(in) :: stem
        integer,     allocatable, intent(in) :: chains(:, :)
        complex(wp), allocatable, intent(in) :: M(:, :, :, :)
        integer,                  intent(in) :: nesc

        character(len=100)          :: filename, history
        character(len=1), parameter :: xyz(-1:1) = [ 'y', 'z', 'x' ]
        character(len=1), parameter :: sph(-1:1) = [ 'm', '0', 'p' ]

#ifdef WITH_COARRAYS
        complex(wp), allocatable :: val(:)[:]
#else
        complex(wp), allocatable :: val(:)
#endif

        integer :: u, itarg, ntarg, ichain, nchain, ipw, npw, icomp, lf, mf, ie, mye

        npw = size(M, 2)
        nchain = size(M, 3)
        ntarg = size(M, 4)

#ifdef WITH_COARRAYS
        allocate (val(nesc)[*])
#else
        allocate (val(nesc))
#endif

        ! there is one output file per target, absorption chain and partial wave
        do itarg = 1, ntarg
            do ichain = 1, nchain
                do ipw = 1, npw

                    val = 0
                    mye = 0
                    do ie = myproc, nesc, nprocs
                        mye = mye + 1
                        val(ie) = M(mye, ipw, ichain, itarg)
                    end do
#ifdef WITH_COARRAYS
                    sync all
                    call co_sum(val, 1)  ! gather coarray from all images to the first one
                    sync all
#endif
                    ! the first image performs the writing
                    if (myproc == 1 .and. any(val /= 0)) then
                        lf = int(sqrt(real(ipw - 1, wp)))
                        mf = ipw - lf*lf - lf - 1
                        history = ''
                        do icomp = 1, size(chains, 1)
                            select case (trim(stem))
                                case ('xyz'); history = xyz(chains(icomp, ichain)) // trim(history)
                                case ('sph'); history = sph(chains(icomp, ichain)) // trim(history)
                            end select
                        end do
                        write (filename, '(3A,I0,A,I0,A,SP,I0,A)') 'rawdip-', trim(history), '-(', itarg, ',', lf, ',', mf, ').txt'
                        open (newunit = u, file = filename, action = 'write', form = 'formatted')
                        write (u, '(2E25.15)') val
                        close (u)
                    end if

                end do
            end do
        end do

    end subroutine write_raw_dipoles


    !> \brief   Write cross sections to a file
    !> \author  J Benda
    !> \date    2020 - 2021
    !>
    !> Write cross sections to a file. The first column of the array is expected to be the photon energy.
    !>
    subroutine write_cross_section (cs, nesc, filename)

        character(len=*),      intent(in) :: filename
        real(wp), allocatable, intent(in) :: cs(:, :)
        integer,               intent(in) :: nesc

#ifdef WITH_COARRAYS
        real(wp), allocatable :: csg(:, :)[:]
#else
        real(wp), allocatable :: csg(:, :)
#endif

        integer :: u, ie, mye, nene, ntgt

        ntgt = size(cs, 1)
        nene = size(cs, 2)

#ifdef WITH_COARRAYS
        ! a local copy of 'cs' for gathering data from parallel images
        allocate (csg(ntgt, nesc)[*])
        csg = 0
        mye = 0
        do ie = myproc, nesc, nprocs
            mye = mye + 1
            csg(:, ie) = cs(:, mye)
        end do

        ! gather coarray from all images to the first one
        sync all
        call co_sum(csg, 1)
        sync all
#else
        ! just copy the array when running in serial mode
        csg = cs
#endif

        ! the first image now writes cross sections to file
        if (myproc == 1 .and. any(csg(2:, :) /= 0)) then
            open (newunit = u, file = filename, action = 'write', form = 'formatted')
            do ie = 1, nesc
                write (u, '(*(E25.15))') csg(:, ie)
            end do
            close (u)
        end if

    end subroutine write_cross_section

end module multidip_io
