! Copyright 2020
!
! For a comprehensive list of the developers that contributed to these codes
! see the UK-AMOR website.
!
! This file is part of UKRmol-out (UKRmol+ suite).
!
!     UKRmol-out is free software: you can redistribute it and/or modify
!     it under the terms of the GNU General Public License as published by
!     the Free Software Foundation, either version 3 of the License, or
!     (at your option) any later version.
!
!     UKRmol-out is distributed in the hope that it will be useful,
!     but WITHOUT ANY WARRANTY; without even the implied warranty of
!     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!     GNU General Public License for more details.
!
!     You should have received a copy of the GNU General Public License
!     along with  UKRmol-out (in source/COPYING). Alternatively, you can also visit
!     <https://www.gnu.org/licenses/>.
!
!> \brief   Main MULTIDIP routines
!> \author  J Benda
!> \date    2020 - 2021
!>
module multidip_routines

    use, intrinsic :: iso_c_binding,   only: c_double, c_int, c_f_pointer
    use, intrinsic :: iso_fortran_env, only: input_unit, int32, int64, real64, output_unit

    ! GBTOlib
    use blas_lapack_gbl,  only: blasint
    use phys_const_gbl,   only: pi, imu, to_eV
    use precisn_gbl,      only: wp

    implicit none

    !> \brief   Intermediate state
    !> \author  J Benda
    !> \date    2020
    !>
    !> Type holding information about a specific irreducible representation of the intermediate state
    !> of a given perturbative order. The states are connected into a linked list to avoid the necessity
    !> of reallocating any large expansion during the calculation. Also, each state is linked to its
    !> "parent" state, which is another instance of this class, which stands on the right-hand side
    !> of the equation for the intermediate state.
    !>
    type IntermediateState

        integer :: order    !< Number of absorbed photons
        integer :: mgvn     !< Irreducible representation of this intermediate state
        integer :: dcomp    !< Dipole operator component responsible for populating this state

        !> Inner region expansion coefficients
        !> - First index: inner region eigenstate index
        !> - Second index: 1 = real part, 2 = imag part
        !> - Third index: scattering energy index
        real(wp), allocatable :: ck(:, :, :)

        !> Outer region expansion coefficients
        !> - First index:  outer region channel index
        !> - Second index: 1 = real part, 2 = imag part
        !> - Third index: scattering energy index
        real(wp), allocatable :: ap(:, :, :)

        !> Photoionisation transition matrix elements. Only used when the "intermediate" state
        !> is actually the final state. In such a case, the expansion coefficients 'ck' and 'ap'
        !> are not allocated.
        !> - First index: outer region channel index
        !> - Second index: 1 = real part, 2 = imag part
        !> - Third index: scattering energy index
        real(wp), allocatable :: dip(:, :, :)

        !> Pointer to the parent intermediate state (if any)
        type(IntermediateState), pointer :: parent => null()

        !> Linked list connections
        type(IntermediateState), pointer :: prev => null()
        type(IntermediateState), pointer :: next => null()

    end type IntermediateState

#ifndef USE_BUILTIN_FINDLOC
    !> \brief   Find element in array
    !> \author  J Benda
    !> \date    2021
    !>
    !> Compatibility replacement for the Fortran 2008 `findloc` function that is missing in GCC < 9.
    !> Only implements the one-dimensional variant for a specific combination of types.
    !>
    interface findloc
        module procedure findloc_32bit_array
        module procedure findloc_64bit_array
    end interface
#endif

contains

    !> \brief   MULTIDIP main subroutine
    !> \authors J Benda
    !> \date    2020 - 2021
    !>
    !> Read the input namelist &mdip from the standard input, call routines for reading the input file, and then call
    !> the main computational routine. Alternatively, when the "--test" switch is present on the command line, it will
    !> run a few unit tests.
    !>
    subroutine multidip_main

        use multidip_io,     only: MolecularData, KMatrix, ScatAkCoeffs, read_wfncoeffs, read_kmatrices, read_molecular_data, &
                                   myproc, nprocs
        use multidip_params, only: nMaxPhotons
        use multidip_tests,  only: run_tests

        type(MolecularData)             :: moldat
        type(KMatrix),      allocatable :: km(:)
        type(ScatAkCoeffs), allocatable :: ak(:)

        character(len=256) :: arg, rmt_data, raw

        integer     :: order, nirr, lusct(8), lukmt(8), nkset(8), iarg, narg, input, i, erange(2)
        logical     :: verbose
        real(wp)    :: omega(nMaxPhotons), first_IP, rasym
        complex(wp) :: polar(3, nMaxPhotons)

        namelist /mdip/ order, lusct, lukmt, nkset, polar, omega, verbose, rmt_data, first_IP, rasym, raw, erange

        order = 1
        lusct = 0
        lukmt = 0
        nkset = 1
        polar = 0
        omega = 0
        rasym = 0
        first_IP = 0
        erange = 0
        verbose = .false.
        rmt_data = 'molecular_data'
        raw = ''

        call print_ukrmol_header(output_unit)

        print '(/,A)', 'Program MULTIDIP: Calculation of multi-photon ionization'

        iarg = 1
        narg = command_argument_count()
        input = input_unit

        do while (iarg <= narg)
            call get_command_argument(iarg, arg)
            if (arg == '--test') then
                call run_tests
                return
            else if (arg == '--input') then
                iarg = iarg + 1
                call get_command_argument(iarg, arg)
                open (input, file = trim(arg), action = 'read', form = 'formatted')
            else if (arg == '--help') then
                print '(/,A,/)', 'Available options:'
                print '(A)', '  --help          display this help and exit'
                print '(A)', '  --input FILE    use FILE as the input file (instead of standard input)'
                print '(A)', '  --test          run internal sanity checks and exit'
                print *
                return
            else
                print '(3A)', 'Unknown command line argument "', trim(arg), '"'
                stop 1
            end if
            iarg = iarg + 1
        end do

        read (input, nml = mdip)

        if (order < 1) then
            print '(A)', 'Order must be positive'
            stop 1
        end if

        if (order > nMaxPhotons) then
            print '(A,I0)', 'Order must be <= ', nMaxPhotons
            stop 1
        end if

        if (raw /= '' .and. raw /= 'sph' .and. raw /= 'xyz' .and. raw /= 'both') then
            print '(A)', 'The parameter "raw" must be empty or one of: "sph", "xyz", "both".'
            stop 1
        end if

        polar(:, order + 1:nMaxPhotons) = 0
        omega(order + 1:nMaxPhotons) = 0

        if (input /= input_unit) close (input)

#ifdef WITH_COARRAYS
        myproc = this_image()
        nprocs = num_images()
        print '(/,A,I0,A,I0)', 'Parallel mode; image ', myproc, ' of ', nprocs
#endif

        print '(/,A,/)', 'Absorbed photons'
        do i = 1, order
            print '(2x,A,I0,3x,A,SP,2F6.3,A,2F6.3,A,2F6.3,A,F0.3,A)', &
                '#', i, 'eX: ', polar(1, i), 'i, eY: ', polar(2, i), 'i, eZ: ', polar(3, i), 'i, energy: ', omega(i) * to_eV, ' eV'
        end do

        if (first_IP > 0) then
            print '(/,A,F0.5,A)', 'Override first IP to ', first_IP * to_eV, ' eV'
        end if

        nirr = count(lukmt > 0)

        allocate (km(nirr))

        if (any(lusct > 0)) then
            allocate (ak(nirr))
            call read_wfncoeffs(ak, lusct)
        end if

        call read_kmatrices(km, lukmt, nkset)
        call read_molecular_data(moldat, trim(rmt_data))

        if (any(erange /= 0)) then
            if (erange(1) < 1 .or. erange(1) > erange(2) .or. erange(2) > minval(km % nescat)) then
                print '(/,A,I0)', 'Error: Energy range must satisfy 1 <= erange(1) <= erange(2) <= ', minval(km % nescat)
            end if
        else
            erange(1) = 1
            erange(2) = minval(km % nescat)
        end if

        if (rasym > moldat % rmatr) then
            if (order <= 2) then
                print '(/,A,F0.1,A,F0.1,A)', 'Interval ', moldat % rmatr, ' to ', rasym, ' will be integrated numerically.'
            else
                print '(/,A)', 'Error: Numerical integration (rasym) is currently only implemented for the second order.'
                stop 1
            end if
        end if

        call multidip_driver(order, moldat, km, ak, omega(1:order), polar(:, 1:order), verbose, first_IP, rasym, raw, erange)

    end subroutine multidip_main


    !> \brief   Central computation routine
    !> \author  J Benda
    !> \date    2020 - 2021
    !>
    !> This subroutine drives the calculation. First it obtains all intermediate states, then it calculates
    !> the final photoionization state, and evaluates the dipole elements and generalized cross sections.
    !> The core of the work is the evaluation of all matrix elements of the type
    !> \f[
    !>     M_{fi,k_n,\dots,k_i} = \langle \Psi_{f}^{(-)} | x_{k_n} \dots \hat{G}^{(+)} x_{k_2} \hat{G}^{(+)} x_{k_1}
    !>             | \Psi_i \rangle
    !> \f]
    !> where \f$ x_j \f$ is the j-th component of the dipole operator, followed by reduction of this tensor with all
    !> provided photon field polarisations \f$ \epsilon_j \f$:
    !> \f[
    !>     M_{fi} = \sum_{k_n,\dots,k_1} \epsilon_{k_n} \dots \epsilon_{k_1} M_{fi,k_n,\dots,k_1}
    !> \f]
    !> The sequence \f$ k_1, k_2, \dots \f$ is referred to as component "history" or "chain" in the code.
    !>
    !> \param[in]  order    Order of the process (= total number of absorbed photons).
    !> \param[in]  moldat   MolecularData object with data read from the file *molecular_data*.
    !> \param[in]  km       KMatrix objects for relevant irreducible representations with data read from RSOLVE K-matrix files.
    !> \param[in]  ak       Wave function coeffs (from RSOLVE) for the same set of irrs as km.
    !> \param[in]  omega    Fixed photon energies in a.u. or zeros for flexible photons.
    !> \param[in]  polar    Photon polarisation vectors or zeros for polarisation averaging.
    !> \param[in]  verbose  Debugging output intensity.
    !> \param[in]  first_IP First ionization potential in a.u. as requested by the input namelist, or zero.
    !> \param[in]  r0       Radius from which to apply asymptotic integrals.
    !> \param[in]  raw      Write raw transition dipoles in spherical (= 'sph') or Cartesian (= 'xyz') basis, or both (= 'both').
    !> \param[in]  erange   Range (subset) of energies in K-matrix files to use for calculations.
    !>
    subroutine multidip_driver (order, moldat, km, ak, omega, polar, verbose, first_IP, r0, raw, erange)

        use multidip_io, only: MolecularData, KMatrix, ScatAkCoeffs, nprocs, get_diptrans

        integer,                         intent(in) :: order, erange(2)
        logical,                         intent(in) :: verbose
        type(MolecularData),             intent(in) :: moldat
        type(KMatrix),      allocatable, intent(in) :: km(:)
        type(ScatAkCoeffs), allocatable, intent(in) :: ak(:)
        real(wp),                        intent(in) :: omega(:), first_IP, r0
        complex(wp),                     intent(in) :: polar(:, :)
        character(len=*),                intent(in) :: raw

        type(IntermediateState), pointer :: states, state

        integer, allocatable :: iidip(:), ifdip(:)

        integer  :: j, s, mgvni, mgvnn, mgvn1, mgvn2, nesc, irri, iki, icomp, nstati, nchani
        real(wp) :: Ei

        iki    = 1                                       ! which K-matrix file (and Ak file) to use for the initial state
        mgvni  = km(iki) % mgvn                          ! IRR of the initial state
        nesc   = km(iki) % nescat                        ! number of scattering energies
        irri   = minloc(abs(moldat % mgvns - mgvni), 1)  ! internal index of initial IRR in molecular_data
        nstati = moldat % mnp1(irri)                     ! number of inner region states in initial IRR
        nchani = moldat % nchan(irri)                    ! number of outer region channels in initial IRR

        Ei = moldat % eig(1, irri)
        if (first_IP > 0) then
            Ei = moldat % etarg(1) - first_IP
        end if

        allocate (states)
        allocate (states % ck(nstati, 2, (nesc + nprocs - 1) / nprocs))
        allocate (states % ap(nchani, 2, (nesc + nprocs - 1) / nprocs))

        ! construct representation of the initial state
        states % order = 0
        states % mgvn = mgvni
        states % dcomp = 0
        states % ck(:, :, :) = 0
        states % ck(1, 1, :) = 1  ! only the first inner state is populated (at all scattering energies)
        states % ap(:, :, :) = 0

        ! for all absorbed photons
        do j = 1, order

            print '(/,2(A,I0))', 'Photon ', j, ' of ', order

            ! for all components of the dipole operator
            do icomp = 1, 3

                call get_diptrans(moldat, icomp, iidip, ifdip)

                ! for all transitions mediated by this component that are stored in molecular_data
                do s = 1, size(iidip)

                    mgvn1 = iidip(s) - 1
                    mgvn2 = ifdip(s) - 1

                    ! apply this transition to all relevant previous intermediate states
                    state => states
                    do while (associated(state))

                        ! process states of previous order
                        if (state % order + 1 == j) then

                            ! IRR of the previous intermediate state (i.e. last element in MGVN history chain)
                            mgvnn = state % mgvn

                            ! skip this transition if it is not applicable to the current IRR or there are no K-matrices for it
                            if ((mgvnn == mgvn1 .or. mgvnn == mgvn2) .and. &
                                any(km(:) % mgvn == mgvn1) .and. &
                                any(km(:) % mgvn == mgvn2)) then

                                ! either calculate the next intermediate state or evaluate the dipole elements
                                if (j < order) then
                                    call solve_intermediate_state(moldat, order, omega, irri, icomp, s, &
                                                                  mgvnn, mgvn1, mgvn2, km, state, verbose, Ei, r0, erange)
                                else
                                    call extract_dipole_elements(moldat, order, omega, irri, icomp, s, &
                                                                 mgvnn, mgvn1, mgvn2, km, ak, state, verbose, Ei, r0, erange)
                                end if

                            end if

                        end if

                        ! move on to the next state
                        state => state % next

                    end do

                end do

            end do

        end do

        ! calculate observables
        call calculate_pw_transition_elements(moldat, order, states, km(iki) % escat(1 : km(iki) % nescat), Ei, omega, polar)
        call calculate_asymmetry_parameters(moldat, order, states, km(iki) % escat(1 : km(iki) % nescat), Ei, omega, raw)

        ! clean up the linked list
        state => states
        do while (associated(state % next))
            state => state % next
            deallocate (state % prev)
        end do
        deallocate (state)

    end subroutine multidip_driver


    !> \brief   Calculate intermediate photoionisation state
    !> \author  J Benda
    !> \date    2020 - 2021
    !>
    !> Solve the intermediate state equation
    !> \f[
    !>     (E_i + j \omega - \hat{H}) \Psi_j = \hat{D} \Psi_{j-1}
    !> \f]
    !> with the right-hand side based on the state provided as argument 'state'.
    !>
    !> \param[in]  moldat   MolecularData object with data read from the file *molecular_data*.
    !> \param[in]  order    Perturbation order of the intermediate state to calculate.
    !> \param[in]  Ephoton  Fixed photon energies in a.u. or zeros for flexible photons.
    !> \param[in]  irri     Index of the previous (parent) state irreducible representation among those present in *molecular_data*.
    !> \param[in]  icomp    Which Cartesian component of the dipole operator will give rise to the intermediate state.
    !> \param[in]  s        Which "transition" in *molecular_data* corresponds to the action of this dipole component on parent.
    !> \param[in]  mgvnn    MGVN of the previous intermediate state.
    !> \param[in]  mgvn1    Ket MGVN for the transition "s" as stored in *molecular_data*.
    !> \param[in]  mgvn2    Bra MGVN for the transition "s" as stored in *molecular_data*.
    !> \param[in]  km       KMatrix objects for relevant irreducible representations with data read from RSOLVE K-matrix files.
    !> \param[inout] state  Tree of intermediate states, pointing at the previous intermediate state to develop into next state.
    !> \param[in]  verbose  Debugging output intensity.
    !> \param[in]  Ei       Initial state energy, either retrieved from RMT molecular data, or override by the input namelist.
    !> \param[in]  r0       Radius from which to apply asymptotic integration.
    !> \param[in]  erange   Range (subset) of energies in K-matrix files to use for calculations.
    !>
    subroutine solve_intermediate_state (moldat, order, Ephoton, irri, icomp, s, mgvnn, mgvn1, mgvn2, &
                                         km, state, verbose, Ei, r0, erange)

        use multidip_io,      only: MolecularData, KMatrix, ScatAkCoeffs, myproc, nprocs, apply_boundary_amplitudes, &
                                    apply_dipole_matrix, scale_boundary_amplitudes
        use multidip_params,  only: closed_interm, closed_range, compt
        use multidip_special, only: coul, solve_complex_system

        type(MolecularData),        intent(in)    :: moldat
        type(KMatrix), allocatable, intent(in)    :: km(:)

        type(IntermediateState), pointer, intent(inout) :: state
        type(IntermediateState), pointer :: last

        integer,  intent(in) :: order, irri, icomp, s, mgvnn, mgvn1, mgvn2, erange(2)
        logical,  intent(in) :: verbose
        real(wp), intent(in) :: Ephoton(:), Ei, r0

        integer  :: nstatn, nstatf, nchann, nchanf, nopen, mgvnf, irrn, irrf, ikn, ikf, ipw, nesc, ie, mye
        real(wp) :: IP, Etotf
        character(len=1) :: transp

        complex(wp), allocatable :: beta(:), app(:), H(:, :), lambda(:), hc(:), hcp(:)

        integer,  allocatable :: lf(:)
        real(wp), allocatable :: Ef(:), fc(:), fcp(:), gc(:), gcp(:), P(:), rhs(:, :), omega(:)
        real(wp), allocatable :: Pw(:, :), wPw(:, :), wPD(:, :), IwPD(:, :), wckp(:, :), Dpsi(:, :, :), corr(:, :), ckp(:, :)

        integer :: i

        if (mgvnn == mgvn1) then
            mgvnf = mgvn2
            transp = 'T'
        else
            mgvnf = mgvn1
            transp = 'N'
        end if

        irrn = findloc(moldat % mgvns, mgvnn, 1)
        irrf = findloc(moldat % mgvns, mgvnf, 1)

        ikn = findloc(km(:) % mgvn, mgvnn, 1)
        ikf = findloc(km(:) % mgvn, mgvnf, 1)

        nstatn = moldat % mnp1(irrn)
        nstatf = moldat % mnp1(irrf)

        nchann = moldat % nchan(irrn)
        nchanf = moldat % nchan(irrf)

        nesc = km(ikf) % nescat
        mye = 0

        allocate (Pw(nstatf, nchanf), wPD(nchanf, 2), IwPD(nchanf, 2), wPw(nchanf, nchanf), wckp(nchanf, 2), omega(order), &
                  Dpsi(nstatf, 2, (nesc + nprocs - 1) / nprocs), fc(nchanf), gc(nchanf), fcp(nchanf), gcp(nchanf), &
                  hc(nchanf), hcp(nchanf), corr(nchanf, 2), &
                  Ef(nchanf), lf(nchanf), ckp(nstatf, 2), H(nchanf, nchanf), lambda(nchanf), beta(nchanf), &
                  app(nchanf))

        print '(2x,A,I0,A,I0,5A)', 'Transition ', mgvnn, ' -> ', mgvnf, ' (', compt(icomp), ', ', transp, ')'

        ! set up a new intermediate state at the end of the storage chain
        last => state
        do while (associated(last % next))
            last => last % next
        end do
        allocate (last % next)
        last % next % prev => last
        last => last % next
        last % order = state % order + 1
        last % mgvn = mgvnf
        last % dcomp = icomp
        last % parent => state
        allocate (last % ck(nstatf, 2, (nesc + nprocs - 1) / nprocs))
        allocate (last % ap(nchanf, 2, (nesc + nprocs - 1) / nprocs))

        ! apply the inner dipoles matrix on the previous inner region solution expansions
        call apply_dipole_matrix(moldat, icomp, s, transp, nstatf, nstatn, state % ck, Dpsi)

        ! for all scattering energies
        do ie = myproc, nesc, nprocs

            ! quantum numbers of the final state
            mye = mye + 1
            IP = km(ikf) % escat(ie) + moldat % etarg(1) - Ei
            call calculate_photon_energies(IP, Ephoton, omega)
            Etotf = Ei + sum(omega(1 : last % order))
            Ef = Etotf - moldat % etarg(moldat % ichl(1:nchanf, irrf))
            lf = moldat % l2p(1:nchanf, irrf)
            nopen = merge(count(Ef + closed_range > 0), count(Ef > 0), closed_interm)

            ! skip this energy if not in the required interval
            if (ie < erange(1) .or. ie > erange(2)) then
                last % ck(:, 1, mye) = 0; last % ck(:, 2, mye) = 0
                last % ap(:, 1, mye) = 0; last % ap(:, 2, mye) = 0
                cycle
            end if

            print '(4x,A,*(I0,A))', 'energy ', ie, ' of ', nesc, &
                                    ' (', count(Ef > 0), ' open / ', nopen, ' open + weakly closed channels of ', nchanf, ')'

            ! evaluate Coulomb functions at this energy for all partial waves in the final IRR
            fc = 0; fcp = 0; gc = 0; gcp = 0; hc = 0; hcp = 0; lambda = 0
            do ipw = 1, nopen
                call coul(lf(ipw), Ef(ipw), moldat % rmatr, fc(ipw), fcp(ipw), gc(ipw), gcp(ipw))
                hc(ipw)  = cmplx(gc(ipw),  fc(ipw),  wp)
                hcp(ipw) = cmplx(gcp(ipw), fcp(ipw), wp)
                lambda(ipw) = hcp(ipw) / hc(ipw)
            end do

            ! inner region part of the right-hand side of the master equation
            rhs = Dpsi(:, :, mye)

            ! outer region correction to the right-hand side
            call multiint(moldat, r0, Ei, omega, mye, last, +1, beta)
            beta = -2 * beta / sqrt(2*abs(Ef))
            where (Ef < 0) beta = beta / imu
            corr(:, 1) = real(beta * (fcp - lambda * fc), wp)
            corr(:, 2) = aimag(beta * (fcp - lambda * fc))
            call apply_boundary_amplitudes(moldat, irrf, 'T', corr, Dpsi(:, :, mye))
            rhs = rhs - 0.5 * Dpsi(:, :, mye)

            ! solve the master equation using the Sherman-Morrison-Woodbury formula
            P = 1 / (Etotf - moldat % eig(1:nstatf, irrf))
            ckp(:, 1) = P * rhs(:, 1)
            ckp(:, 2) = P * rhs(:, 2)

            if (nopen > 0) then

                ! compose the reduced inverse Hamiltonian
                H = 0
                call scale_boundary_amplitudes(moldat, irrf, P, Pw)
                call apply_boundary_amplitudes(moldat, irrf, 'N', Pw, wPw)
                H(1:nopen, 1:nopen) = wPw(1:nopen, 1:nopen)
                do i = 1, nopen
                    H(i, i) = H(i, i) + 2/lambda(i)
                end do
                do i = nopen + 1, nchanf
                    H(i, i) = 1
                end do
                call apply_boundary_amplitudes(moldat, irrf, 'N', ckp, wPD)

                ! solve the reduced system
                IwPD = 0
                call solve_complex_system(nopen, H, wPD, IwPD)

                ! expand the reduced solution to the full solution
                call apply_boundary_amplitudes(moldat, irrf, 'T', IwPD, ckp)
                ckp(:, 1) = P * (rhs(:, 1) - ckp(:, 1))
                ckp(:, 2) = P * (rhs(:, 2) - ckp(:, 2))

            end if

            ! extract the outer region channel amplitudes
            call apply_boundary_amplitudes(moldat, irrf, 'N', ckp, wckp)
            app = cmplx(wckp(:, 1), wckp(:, 2), wp)
            where (hc /= 0)
                app = (app - beta*fc) / hc
            elsewhere
                app = 0
            end where

            if (verbose) print '(4x,A,*(E25.15))', ' - beta: ', beta(1:nopen)
            if (verbose) print '(4x,A,*(E25.15))', ' - ap:   ', app(1:nopen)

            last % ck(:, 1, mye) = ckp(:, 1)
            last % ck(:, 2, mye) = ckp(:, 2)

            last % ap(:, 1, mye) = real(app, wp)
            last % ap(:, 2, mye) = aimag(app)

        end do

    end subroutine solve_intermediate_state


    !> \brief   Calculate dipole elements from intermediate and final states
    !> \author  J Benda
    !> \date    2020 - 2021
    !>
    !> Calculates the transition dipole matrix element between the last intermediate state and the final
    !> stationary photoionization state.
    !>
    !> \param[in]  moldat   MolecularData object with data read from the file *molecular_data*.
    !> \param[in]  order    Perturbation order of the intermediate state to calculate.
    !> \param[in]  Ephoton  Fixed photon energies in a.u. or zeros for flexible photons.
    !> \param[in]  irri     Index of the previous (parent) state irreducible representation among those present in *molecular_data*.
    !> \param[in]  icomp    Which Cartesian component of the dipole operator will give rise to the intermediate state.
    !> \param[in]  s        Which "transition" in *molecular_data* corresponds to the action of this dipole component on parent.
    !> \param[in]  mgvnn    MGVN of the previous intermediate state.
    !> \param[in]  mgvn1    Ket MGVN for the transition "s" as stored in *molecular_data*.
    !> \param[in]  mgvn2    Bra MGVN for the transition "s" as stored in *molecular_data*.
    !> \param[in]  km       KMatrix objects for relevant irreducible representations with data read from RSOLVE K-matrix files.
    !> \param[in]  ak       Wave function coeffs (from RSOLVE) for the same set of irrs as km.
    !> \param[inout] state  Tree of intermediate states, pointing at the previous intermediate state to develop into next state.
    !> \param[in]  verbose  Debugging output intensity.
    !> \param[in]  Ei       Initial state energy, either retrieved from RMT molecular data, or override by the input namelist.
    !> \param[in]  r0       Radius from which to apply asymptotic integrals.
    !> \param[in]  erange   Range (subset) of energies in K-matrix files to use for calculations.
    !>
    subroutine extract_dipole_elements (moldat, order, Ephoton, irri, icomp, s, mgvnn, mgvn1, mgvn2, &
                                        km, ak, state, verbose, Ei, r0, erange)

        use multidip_io,      only: MolecularData, KMatrix, ScatAkCoeffs, myproc, nprocs, apply_dipole_matrix
        use multidip_params,  only: rone, rzero, cone, czero, compt, maxtarg
        use multidip_special, only: scatAk,calculate_S_matrix, blas_dgemv => dgemv, blas_zgemv => zgemv

        type(MolecularData),             intent(in)    :: moldat
        type(KMatrix),      allocatable, intent(in)    :: km(:)
        type(ScatAkCoeffs), allocatable, intent(in)    :: ak(:)

        type(IntermediateState), pointer, intent(inout) :: state
        type(IntermediateState), pointer                :: last

        integer,  intent(in) :: irri, order, icomp, s, mgvnn, mgvn1, mgvn2, erange(2)
        logical,  intent(in) :: verbose
        real(wp), intent(in) :: Ephoton(:), Ei, r0

        character(len=1) :: transp

        complex(wp), allocatable :: Af(:, :), dm(:), dp(:), Sm(:, :)
        real(wp), allocatable :: Dpsi(:, :, :), Ef(:), echlf(:), d_inner(:, :), d_outer(:, :), Re_Af(:, :), Im_Af(:, :)
        real(wp), allocatable :: omega(:), d_total(:, :), kmat(:, :)
        integer,  allocatable :: lf(:)

        real(wp) :: Etotf, IP
        integer  :: mgvnf, irrn, irrf, ikn, ikf, nstatn, nstatf, nchann, nchanf, nesc, ie, nopen, nochf, mye, nene, maxchan, i

        integer(blasint) :: m, n, lds, inc = 1

        if (mgvnn == mgvn1) then
            mgvnf = mgvn2
            transp = 'T'
        else
            mgvnf = mgvn1
            transp = 'N'
        end if

        irrn = findloc(moldat % mgvns, mgvnn, 1)
        irrf = findloc(moldat % mgvns, mgvnf, 1)

        ikn = findloc(km(:) % mgvn, mgvnn, 1)
        ikf = findloc(km(:) % mgvn, mgvnf, 1)

        nstatn = moldat % mnp1(irrn)
        nstatf = moldat % mnp1(irrf)

        nchann = km(ikn) % nchan
        nchanf = km(ikf) % nchan

        nesc = km(ikf) % nescat
        nene = (nesc + nprocs - 1) / nprocs
        mye = 0

        maxchan = nchanf
        if (maxtarg > 0) then
            maxchan = count(moldat % ichl(1:nchanf, irrf) <= maxtarg)
        end if

        allocate (Dpsi(nstatf, 2, nene), omega(order), Ef(nchanf), lf(nchanf), echlf(nchanf), &
                  Re_Af(nstatf, maxchan), Im_Af(nstatf, maxchan), Af(nstatf, maxchan), Sm(nchanf, nchanf), &
                  d_inner(maxchan, 2), d_outer(maxchan, 2), d_total(maxchan, 2), dm(maxchan), dp(nchanf))

        print '(2x,A,I0,A,I0,5A)', 'Transition ', mgvnn, ' -> ', mgvnf, ' (', compt(icomp), ', ', transp, ')'

        ! set up a final state at the end of the storage chain
        last => state
        do while (associated(last % next))
            last => last % next
        end do
        allocate (last % next)
        last % next % prev => last
        last => last % next
        last % order = order
        last % mgvn = mgvnf
        last % dcomp = icomp
        last % parent => state
        allocate (last % dip(maxchan, 2, nene))

        ! apply the inner dipoles matrix on the previous inner region solution expansions
        call apply_dipole_matrix(moldat, icomp, s, transp, nstatf, nstatn, state % ck, Dpsi)

        ! position K-matrix file pointer at the beginning of (this process') K-matrix data
        call km(ikf) % reset_kmatrix_position(skip = myproc - 1)
        if (allocated(ak)) call ak(ikf) % reset_wfncoeffs_position(skip = myproc - 1)

        ! for all scattering energies
        do ie = myproc, nesc, nprocs

            mye = mye + 1
            last % dip(:, :, mye) = 0

            ! read the K-matrix from the file and then skip the next 'nprocs - 1' records used by other images
            if (ie > erange(2)) cycle
            call km(ikf) % get_kmatrix(kmat, skip = nprocs - 1)
            if (allocated(ak)) call ak(ikf) % get_wfncoeffs(Re_Af, Im_Af, skip = nprocs - 1)
            if (ie < erange(1)) cycle

            ! quantum numbers of the final state
            Etotf = km(ikf) % escat(ie) + moldat % etarg(1)
            IP = Etotf - Ei
            call calculate_photon_energies(IP, Ephoton, omega)
            echlf = moldat % etarg(moldat % ichl(1:nchanf, irrf)) - moldat % etarg(1)
            Ef = km(ikf) % escat(ie) - echlf        ! photoelectron energies in all (open and closed) channels
            lf = moldat % l2p(1:nchanf, irrf)       ! photoelectron angular momentum in all (open and closed) channels
            nopen = count(Ef > 0)                   ! number of all open channels
            nochf = min(nopen, maxchan)             ! number of open channels for which we want to obtain dipoles

            print '(4x,A,*(I0,A))', 'energy ', ie, ' of ', nesc, ' (', nopen, ' open channels of ', nchanf, ')'

            ! obtain wave function coefficients in the final IRR from the Ak file data structure, or calculate them on the fly
            if (allocated(ak)) then
                !Re_Af(:, 1:nochf) = ak(ikf) % ReA(:, 1:nochf, mye)
                !Im_Af(:, 1:nochf) = ak(ikf) % ImA(:, 1:nochf, mye)
            else
                call scatAk(moldat % rmatr, &
                           moldat % eig(1:nstatf, irrf), &
                           Etotf, &
                           moldat % wamp(irrf) % mat(:, :, 1), &
                           Ef, &
                           lf, &
                           kmat, &
                           Af(:, 1:nochf))

                Re_Af(:, 1:nochf) = real(Af(:, 1:nochf), wp)
                Im_Af(:, 1:nochf) = aimag(Af(:, 1:nochf))
            end if

            ! contract D*psi with complex-conjugated wave function coefficients Ak
            m = nstatf
            n = nochf
            d_inner = 0
            call blas_dgemv('T', m, n,  rone, Re_Af, m, Dpsi(:, 1, mye), inc, rzero, d_inner(:, 1), inc)
            call blas_dgemv('T', m, n, +rone, Im_Af, m, Dpsi(:, 2, mye), inc, +rone, d_inner(:, 1), inc)
            call blas_dgemv('T', m, n,  rone, Re_Af, m, Dpsi(:, 2, mye), inc, rzero, d_inner(:, 2), inc)
            call blas_dgemv('T', m, n, -rone, Im_Af, m, Dpsi(:, 1, mye), inc, +rone, d_inner(:, 2), inc)

            ! outer region correction
            dm = 0
            dp = 0
            call multiint(moldat, r0, Ei, omega, mye, last, -1, dm(1:nochf))
            call multiint(moldat, r0, Ei, omega, mye, last, +1, dp(1:nopen))
            dm(1:nochf) = dm(1:nochf) * imu / sqrt(2*pi*sqrt(2*Ef(1:nochf)))
            dp(1:nopen) = dp(1:nopen) * imu / sqrt(2*pi*sqrt(2*Ef(1:nopen)))
            if (verbose) print '(5x,A,1x,*(1x,E25.15))', 'dm = ', dm(1:nochf)
            if (verbose) print '(5x,A,1x,*(1x,E25.15))', 'dp = ', dp(1:nopen)
            if (any(dp /= 0)) then
                m = nopen; n = nochf; lds = nchanf
                call calculate_S_matrix(kmat, Sm, nopen)
                call blas_zgemv('T', m, n, -cone, Sm, lds, dp, inc, cone, dm, inc)  ! dm = dm - S^T dp
            end if
            d_outer(:, 1) = real(dm, wp)
            d_outer(:, 2) = aimag(dm)

            ! combine and store the partial wave dipoles
            d_total = d_inner + d_outer
            last % dip(:, :, mye) = d_total

            if (verbose) print '(5x,A,1x,*(1x,E25.15))', 'd_inner = ', (d_inner(n, 1), d_inner(n, 2), n = 1, nochf)
            if (verbose) print '(5x,A,1x,*(1x,E25.15))', 'd_outer = ', (d_outer(n, 1), d_outer(n, 2), n = 1, nochf)
            if (verbose) print '(5x,A,1x,*(1x,E25.15))', 'd_total = ', (d_total(n, 1), d_total(n, 2), n = 1, nochf)

        end do

    end subroutine extract_dipole_elements


    !> \brief   Calculate energy of each photon
    !> \author  J Benda
    !> \date
    !>
    !> Calculate energies of all photons. The positive elements of the input array `Ephoton` specify fixed
    !> energies of selected photons. The remaining energy needed to reach the ionisation potential is divided
    !> among the remaining photons.
    !>
    !> \param[in]    IP       Summed energy of all photons.
    !> \param[in]    Ephoton  User-specified fixed energies of photons or zeros for flexible photons.
    !> \param[inout] omega    Calculated energies of all photons.
    !>
    subroutine calculate_photon_energies (IP, Ephoton, omega)

        real(wp),              intent(in)    :: IP, Ephoton(:)
        real(wp), allocatable, intent(inout) :: omega(:)

        where (Ephoton /= 0)
            omega = Ephoton
        elsewhere
            omega = (IP - sum(Ephoton)) / (size(omega) - count(Ephoton /= 0))
        end where

    end subroutine calculate_photon_energies


    !> \brief   Evaluate the correction dipole integral for all orders
    !> \author  J Benda
    !> \date    2020 - 2021
    !>
    !> If the parent intermediate state has a non-vanishing outer region part, integrate the final wave function with it
    !> and multiply by the associated expansion coefficients 'ap'. Then, iteratively, for all parent states of the parent
    !> state, add their contribution by means of the Coulomb-Green's function (multiplied by their expansion coefficients).
    !> The deeper in the absorption chain we get, the more dimensions the resulting Coulomb-Green's integral has.
    !>
    !> \param[in]  moldat  multidip_io::MolecularData object with data read from *molecular_data*.
    !> \param[in]  r0      Radius from which to apply asymptotic integrals.
    !> \param[in]  Ei      Total energy of the initial state.
    !> \param[in]  omega   Energy of all photons.
    !> \param[in]  ie      Local energy index.
    !> \param[in]  state   Tree of parent intermediate states pointed to the last intermediate state.
    !> \param[in]  sb      Kind (sign) of the outer-most Coulomb-Hankel function.
    !> \param[out] dip     Evaluated multi-photon matrix element.
    !>
    subroutine multiint (moldat, r0, Ei, omega, ie, state, sb, dip)

        use multidip_io,     only: MolecularData
        use multidip_params, only: closed_interm, closed_range

        type(MolecularData),              intent(in) :: moldat
        type(IntermediateState), pointer, intent(in) :: state

        complex(wp),              intent(inout) :: dip(:)
        real(wp),    allocatable, intent(inout) :: omega(:)
        real(wp),                 intent(in)    :: Ei, r0
        integer,                  intent(in)    :: ie, sb

        complex(wp), allocatable :: k(:)
        integer,     allocatable :: l(:), m(:)

        real(wp)    :: c, a, Ekf, Etot
        integer     :: nphot, mgvn, ichan, irr

        dip = 0

        if (.not. associated(state % parent)) stop 1    ! state with no parent should not be a final state in the first place

        a = moldat % rmatr                              ! R-matrix radius
        c = 0                                           ! dipole damping factor
        nphot = state % order                           ! number of photons absorbed by the final state
        Etot = Ei + sum(omega(1:nphot))                 ! total energy of the final state
        mgvn = state % mgvn                             ! MGVN of the final state
        irr = findloc(moldat % mgvns, mgvn, 1)          ! index of this spin-symmetry in molecular_data

        allocate (k(nphot), l(nphot), m(nphot - 1))

        do ichan = 1, size(dip)

            ! set up quantum numbers for this channel
            Ekf = Etot - moldat % etarg(moldat % ichl(ichan, irr))  ! always positive for final states (but not for intermediate)
            if (Ekf + closed_range <= 0) cycle                      ! ignore strongly closed channels
            if (Ekf < 0 .and. .not. closed_interm) cycle            ! optionally ignore also weakly closed channels
            k(1) = sqrt(2*abs(Ekf)); if (Ekf < 0) k(1) = k(1) * imu
            l(1) = moldat % l2p(ichan, irr)

            ! evaluate the correction dipole integral for this final channel
            dip(ichan) = multiint_chain(moldat, r0, Ei, omega, ie, c, 1, state, ichan, sb, k, l, m)

        end do

    end subroutine multiint


    !> \brief   Calculate dipole correction integrals at given absorption depth
    !> \author  J Benda
    !> \date    2020 - 2021
    !>
    !> Recursively evaluates the outer region contributions multi-photon transition element contributions from all combinations
    !> of channels at all absorption levels that share the initial sequence given by k, l and m.
    !>
    !> \param[in]  moldat  multidip_io::MolecularData object with data read from *molecular_data*.
    !> \param[in]  r0      Radius from which to apply asymptotic integrals.
    !> \param[in]  Ei      Total energy of the initial state.
    !> \param[in]  omega   Energy of all photons.
    !> \param[in]  ie      Local energy index.
    !> \param[in]  c       Dipole exponential damping coefficient.
    !> \param[in]  N       Index of the dipole operator to consider on this recursion level.
    !> \param[in]  state   Tree of parent intermediate states pointed to the last intermediate state.
    !> \param[in]  ichanf  Final channel after action of the dipole operator at this recursion level.
    !> \param[in]  sb      Kind (sign) of the outer-most Coulomb-Hankel function.
    !> \param[inout]  k    Linear momenta of the final and initial dipole transition channels for at previous recursion levels.
    !> \param[inout]  l    Angular momenta of the final and initial dipole transition channels for at previous recursion levels.
    !> \param[inout]  m    Angular projections of the final and initial dipole transition channels for at previous recursion levels.
    !>
    !> \return integ Evaluated multi-photon matrix element.
    !>
    !> \todo We do not need to calculate the integrals for all final-initial channel pairs as done now, because when the initial
    !>       channels have the same quantum number 'm', the integrals will differ only by the coefficient 'ap'. This would reduce
    !>       the run time for calculations with large partial wave expansion.
    !>
    recursive complex(wp) function multiint_chain (moldat, r0, Ei, omega, ie, c, N, state, ichanf, sb, k, l, m) result (integ)

        use multidip_integ,  only: nested_cgreen_integ
        use multidip_io,     only: MolecularData
        use multidip_params, only: closed_interm, closed_range

        type(MolecularData),              intent(in) :: moldat
        type(IntermediateState), pointer, intent(in) :: state
        type(IntermediateState), pointer             :: parent

        complex(wp), allocatable, intent(inout) :: k(:)
        real(wp),    allocatable, intent(inout) :: omega(:)
        integer,     allocatable, intent(inout) :: l(:), m(:)
        real(wp),                 intent(in)    :: Ei, r0
        integer,                  intent(in)    :: ie, sb, N, ichanf

        complex(wp) :: ap, kr(N + 1)
        real(wp)    :: c, a, Ek, Qion, Qpws, Etot
        integer     :: mgvni, mgvnf, nchani, irri, ichani, nphot, dcomp, lr(N + 1), mr(N)

        integ = 0

        if (state % parent % order == 0) return         ! if the parent is the initial state, there are not tails to integrate

        a = moldat % rmatr                              ! R-matrix radius
        dcomp = state % dcomp                           ! last polarisation component absorbed by the final state
        parent => state % parent                        ! previous intermediate state
        nphot = parent % order                          ! number of photons absorbed by the previous intermediate state
        mgvnf = state % mgvn                            ! irreducible representation of the current state
        mgvni = parent % mgvn                           ! irreducible representation of the previos state
        irri = findloc(moldat % mgvns, mgvni, 1)        ! index of this spin-symmetry in molecular_data
        nchani = moldat % nchan(irri)                   ! number of channels in the previous state irreducible representation
        Etot = Ei + sum(omega(1:nphot))                 ! total energy of the previous intermediate state

        ! loop over all channels of the intermediate state that are dipole-coupled to ichanf
        do ichani = 1, nchani

            ! set up quantum numbers for this channel
            Ek = Etot - moldat % etarg(moldat % ichl(ichani, irri))     ! can be negative
            if (Ek + closed_range <= 0) cycle                           ! ignore strongly closed channels
            if (Ek < 0 .and. .not. closed_interm) cycle                 ! optionally ignore also weakly closed channels
            k(N + 1) = sqrt(2*abs(Ek)); if (Ek < 0) k(N + 1) = k(N + 1) * imu
            l(N + 1) = moldat % l2p(ichani, irri)
            kr(1 : N + 1) = k(N + 1 : 1 : -1)
            lr(1 : N + 1) = l(N + 1 : 1 : -1)

            ! dipole angular integrals
            Qion = channel_coupling_ion(moldat, dcomp, mgvnf, mgvni, ichanf, ichani)
            Qpws = channel_coupling_pws(moldat, dcomp, mgvnf, mgvni, ichanf, ichani)

            ! outer region expansion coefficient for this partial wave and energy
            ap = cmplx(parent % ap(ichani, 1, ie), parent % ap(ichani, 2, ie), wp)

            ! evaluate Coulomb-Green's integrals for dipole-coupled ions
            if (Qion /= 0) then
                m(N) = 0
                mr(1:N) = m(N:1:-1)
                if (ap /= 0) integ = integ + Qion * ap * nested_cgreen_integ(a, r0, c, N, +1, sb, mr, lr, kr)
                integ = integ + Qion * multiint_chain(moldat, r0, Ei, omega, ie, c, N + 1, parent, ichani, sb, k, l, m)
            end if

            ! evaluate Coulomb-Green's integrals for dipole-coupled partial waves
            if (Qpws /= 0) then
                m(N) = 1
                mr(1:N) = m(N:1:-1)
                if (ap /= 0) integ = integ + Qpws * ap * nested_cgreen_integ(a, r0, c, N, +1, sb, mr, lr, kr)
                integ = integ + Qpws * multiint_chain(moldat, r0, Ei, omega, ie, c, N + 1, parent, ichani, sb, k, l, m)
            end if

        end do

    end function multiint_chain


    !> \brief   Ion channel dipole coupling
    !> \author  J Benda
    !> \date    2020
    !>
    !> Returns the ion transition dipole element between the channels. This is diagonal in
    !> quantum numbers of the partial waves and simply equal to the corresponding N-electron
    !> propery integral.
    !>
    !> \param[in]  moldat  multidip_io::MolecularData object with data read from *molecular_data*.
    !> \param[in]  dcomp   Index of the Cartesian component of the dipole operator.
    !> \param[in]  mgvnf   MGVN of the final state channel.
    !> \param[in]  mgvni   MGVN of the initial state channel.
    !> \param[in]  ichanf  Index of the final state channel.
    !> \param[in]  ichani  Index of the initial state channel.
    !>
    real(wp) function channel_coupling_ion (moldat, dcomp, mgvnf, mgvni, ichanf, ichani) result (Qion)

        use multidip_io, only: MolecularData

        type(MolecularData), intent(in) :: moldat
        integer,             intent(in) :: dcomp, mgvnf, mgvni, ichanf, ichani

        integer :: tr(3) = [ 3, 1, 2 ]  ! how crlv is stored: x is last, y first, z in between
        integer :: irri, irrf, itargi, itargf, li, lf, mi, mf

        irrf = findloc(moldat % mgvns, mgvnf, 1)
        irri = findloc(moldat % mgvns, mgvni, 1)

        lf = moldat % l2p(ichanf, irrf); mf = moldat % m2p(ichanf, irrf); itargf = moldat % ichl(ichanf, irrf)
        li = moldat % l2p(ichani, irri); mi = moldat % m2p(ichani, irri); itargi = moldat % ichl(ichani, irri)

        if (lf /= li .or. mf /= mi) then
            Qion = 0
        else
            Qion = moldat % crlv(itargf, itargi, tr(dcomp))
        end if

    end function channel_coupling_ion


    !> \brief   Partial wave channel dipole coupling
    !> \author  J Benda
    !> \date    2020
    !>
    !> Returns the partial wave transition dipole element between the channels. This is
    !> diagonal in the ion states and proportional to the Gaunt coefficient.
    !>
    !> \param[in]  moldat  multidip_io::MolecularData object with data read from *molecular_data*.
    !> \param[in]  dcomp   Index of the Cartesian component of the dipole operator.
    !> \param[in]  mgvnf   MGVN of the final state channel.
    !> \param[in]  mgvni   MGVN of the initial state channel.
    !> \param[in]  ichanf  Index of the final state channel.
    !> \param[in]  ichani  Index of the initial state channel.
    !>
    real(wp) function channel_coupling_pws (moldat, dcomp, mgvnf, mgvni, ichanf, ichani) result (Qpws)

        use multidip_io, only: MolecularData

        type(MolecularData), intent(in) :: moldat
        integer,             intent(in) :: dcomp, mgvnf, mgvni, ichanf, ichani

        integer :: m(3) = [ +1, -1, 0 ]  ! real spherical harmonics Y1m associated with directions x,y,z
        integer :: irri, irrf, itargi, itargf, li, lf, mi, mf, ipwi, ipwf

        irrf = findloc(moldat % mgvns, mgvnf, 1)
        irri = findloc(moldat % mgvns, mgvni, 1)

        itargf = moldat % ichl(ichanf, irrf)
        itargi = moldat % ichl(ichani, irri)

        if (itargf /= itargi) then
            Qpws = 0
        else
            lf = moldat % l2p(ichanf, irrf); mf = moldat % m2p(ichanf, irrf); ipwf = lf*(lf + 1) + mf + 1
            li = moldat % l2p(ichani, irri); mi = moldat % m2p(ichani, irri); ipwi = li*(li + 1) + mi + 1
            Qpws = sqrt(4*pi/3) * moldat % gaunt(ipwf, ipwi, m(dcomp))
        end if

    end function channel_coupling_pws


    !> \brief   Calculate partial wave dipoles, oriented dipoles and cross sections
    !> \author  J Benda
    !> \date    2020 - 2021
    !>
    !> Given the uncontracted partial wave dipoles
    !> \f[
    !>     M_{i_f, l_f, m_f, j_1, \dots, j_n}^{(n)}
    !> \f]
    !> calculated in \ref extract_dipole_elements, where \f$ i_f \f$ is the index of the final ion state, \f$ l_f \f$ and
    !> \f$ m_f \f$ denote the emission partial wave and \f$ j_1, \dots, j_n \f$ are the indices of components of the polarisation
    !> vectors, evaluate the partial wave transition matrix elements
    !> \f[
    !>     M_{i_f, l_f, m_f} = \mathrm{i}^{-l_f} \mathrm{e}^{\mathrm{i} \sigma_f} \sum_{j_1, \dots, j_n}
    !>                         \epsilon_{j_1} \dots \epsilon_{j_n} M_{i_f, l_f, mf_, j_1, \dots, j_n}^{(n)}
    !> \f]
    !> contracted with the polarisation vectors themselves, and the fixed-orientation generalized cross section
    !> \f[
    !>    \sigma_f^{(n)} = 2\pi (2\pi\alpha\omega)^n \sum_{l_f m_f} \left| M_{i_f, l_f, m_f} \right|^2
    !> \f]
    !>
    !> \param[in]  moldat   multidip_io::MolecularData object with data read from the file *molecular_data*.
    !> \param[in]  order    Perturbation order matrix elements (= number of absorbed photons).
    !> \param[in]  state    Tree of intermediate and final states.
    !> \param[in]  escat    Scattering energies in a.u., as stored in the K-matrix files.
    !> \param[in]  Ei       Total energy of the initial state.
    !> \param[in]  Ephoton  Fixed photon energies in a.u. or zeros for flexible photons.
    !> \param[in]  polar    Photon polarisations or zeros for polarisation averaging.
    !>
    subroutine calculate_pw_transition_elements (moldat, order, state, escat, Ei, Ephoton, polar)

        use multidip_io,      only: MolecularData, myproc, nprocs, write_partial_wave_moments, write_cross_section
        use multidip_params,  only: alpha, maxtarg
        use multidip_special, only: cphase

        type(MolecularData),              intent(in) :: moldat
        type(IntermediateState), pointer, intent(in) :: state
        type(IntermediateState), pointer             :: ptr, parent

        integer,     intent(in) :: order
        real(wp),    intent(in) :: escat(:)
        complex(wp), intent(in) :: polar(:, :)
        real(wp),    intent(in) :: Ei, Ephoton(:)
        integer                 :: i, nesc, ntarg, nchan, mxchan, lf, ichan, itarg, ie, irr, mgvn, mye, nirr, nene
        real(wp)                :: IP, Ef, kf, sigma, Q
        complex(wp)             :: d, ej
        complex(wp), allocatable :: M(:, :, :)
        real(wp),    allocatable :: cs(:, :), omega(:)

        nirr = size(moldat % mgvns)             ! number of irreducible representations in molecular_data
        nesc = size(escat)                      ! total number of scattering energies
        mxchan = maxval(moldat % nchan)         ! maximal number of channels per irreducible representation
        ntarg = size(moldat % etarg)            ! number of targes (residual ions)
        nene = (nesc + nprocs - 1) / nprocs     ! number of scattering energies managed by this image

        ! if not all final targets are required, reduce some of the above limits
        if (maxtarg > 0) then
            mxchan = 0
            do irr = 1, nirr
                nchan = moldat % nchan(irr)
                mxchan = max(mxchan, count(moldat % ichl(1:nchan, irr) <= maxtarg))
            end do
            ntarg = min(ntarg, maxtarg)
        end if

        ! set up final storage
        allocate (omega(order), M(mxchan, nene, 0:7), cs(2 + ntarg, nene))
        M = 0
        cs = 0

        ! the dipole elements stored in molecular_data do not contain the charge, so for each transition we need a factor of -1
        Q = (-1)**order

        ! calculate fixed-in-space transition matrix elements
        ptr => state
        do while (associated(ptr))
            if (ptr % order == order) then

                ! get polarisation factor for this absorption chain
                ej = polar(ptr % dcomp, ptr % order)
                parent => ptr % parent
                do while (associated(parent % parent))
                    ej = ej * polar(parent % dcomp, parent % order)
                    parent => parent % parent
                end do

                ! add transition element contribution from this absorption chain
                mye = 0
                do ie = myproc, nesc, nprocs
                    mye = mye + 1
                    do ichan = 1, size(ptr % dip, 1)
                        d = ej * cmplx(ptr % dip(ichan, 1, mye), ptr % dip(ichan, 2, mye), wp)
                        M(ichan, mye, ptr % mgvn) = M(ichan, mye, ptr % mgvn) + Q * d
                    end do
                end do

            end if
            ptr => ptr % next
        end do

        ! write partial wave transition moments without Coulomb phase factor
        call write_partial_wave_moments(moldat, M, nesc, '')

        ! process transition matrix elements
        mye = 0
        do ie = myproc, nesc, nprocs

            mye = mye + 1
            IP = escat(ie) + moldat % etarg(1) - Ei
            call calculate_photon_energies(IP, Ephoton, omega)

            ! add phase factors to transition matrix elements
            do irr = 1, nirr
                mgvn = moldat % mgvns(irr)
                nchan = min(int(moldat % nchan(irr)), mxchan)
                do ichan = 1, nchan
                    lf = moldat % l2p(ichan, irr)
                    Ef = escat(ie) - moldat % etarg(moldat % ichl(ichan, irr)) + moldat % etarg(1)
                    if (Ef > 0) then
                        kf = sqrt(2*Ef)
                        sigma = cphase(lf, kf)
                        M(ichan, mye, mgvn) = M(ichan, mye, mgvn) * imu**(-lf) * (cos(sigma) + imu*sin(sigma))
                    end if
                end do
            end do

            ! calculate oriented cross section
            cs(1, mye) = omega(1) * to_eV
            do itarg = 1, ntarg
                cs(2 + itarg, mye) = 0
                do irr = 1, nirr
                    mgvn = moldat % mgvns(irr)
                    nchan = min(int(moldat % nchan(irr)), mxchan)
                    do ichan = 1, nchan
                        if (itarg == moldat % ichl(ichan, irr)) then
                            d = M(ichan, mye, mgvn)
                            cs(2 + itarg, mye) = cs(2 + itarg, mye) + d * conjg(d)
                        end if
                    end do
                end do
            end do
            cs(2, mye) = sum(cs(3:, mye))
            cs(2:, mye) = cs(2:, mye) * 2*pi * (2*pi*alpha)**order * product(omega)

        end do

        ! write partial wave transition moments with Coulomb phase factor
        call write_partial_wave_moments(moldat, M, nesc, '+cphase')

        ! write fixed-in-space cross section
        call write_cross_section(cs, nesc, 'gen_photo_xsec.txt')

    end subroutine calculate_pw_transition_elements


    !> \brief   Calculate cross sections and asymmetry parameters
    !> \author  J Benda
    !> \date    2020 - 2021
    !>
    !> Evaluate coefficients \f$ \beta_L \f$ in the expansion of differential cross section averaged over
    !> molecular orientation:
    !> \f[
    !>     \frac{\mathrm{d}\sigma}{\mathrm{d}\Omega} = \frac{\beta_0}{4\pi} \left(1+\sum_{L=1}^{2n} \beta_L P_L(\cos\theta) \right).
    !> \f]
    !> The quantity \f$ \beta_0 \f$ is equal to the integral cross section.
    !>
    !> \param[in]  moldat   multidip_io::MolecularData object with data read from the file *molecular_data*.
    !> \param[in]  order    Perturbation order matrix elements (= number of absorbed photons).
    !> \param[in]  state    Tree of intermediate and final states.
    !> \param[in]  escat    Scattering energies in a.u., as stored in the K-matrix files.
    !> \param[in]  Ei       Total energy of the initial state.
    !> \param[in]  Ephoton  Fixed photon energies in a.u. or zeros for flexible photons.
    !> \param[in]  raw      Write raw transition dipoles (in spherical or cartesian basis).
    !>
    subroutine calculate_asymmetry_parameters (moldat, order, state, escat, Ei, Ephoton, raw)

        use dipelm_special_functions, only: sph_basis_transform_elm
        use multidip_io,              only: MolecularData, myproc, nprocs, write_raw_dipoles, write_cross_section
        use multidip_params,          only: alpha, maxtarg
        use multidip_special,         only: cphase, beta_contraction_tensor

        character(len=*),                 intent(in) :: raw
        type(MolecularData),              intent(in) :: moldat
        type(IntermediateState), pointer, intent(in) :: state
        type(IntermediateState), pointer             :: ptr, parent

        integer,     intent(in) :: order
        real(wp),    intent(in) :: escat(:)
        real(wp),    intent(in) :: Ei, Ephoton(:)

        integer                  :: nesc, i, ii, ij, li, lj, mi, mj, nchain, ichain, jchain, qi(order), qj(order), p(order)
        integer                  :: ntarg, ichan, gchan, ie, mye, irr, L, nene, maxl, pw, pwi, pwj
        integer,     allocatable :: chains(:, :)
#ifdef WITH_COARRAYS
        real(wp),    allocatable :: T(:, :, :, :)[:]
#else
        real(wp),    allocatable :: T(:, :, :, :)
#endif
        real(wp),    allocatable :: omega(:), cs(:, :), beta(:, :)
        complex(wp), allocatable :: M_xyz(:, :, :, :), M_sph(:, :, :, :)
        complex(wp)              :: d, cpl
        real(wp)                 :: Ek, k, sigma, IP, Q
        character(len=50)        :: filename

        nesc = size(escat)                      ! total number of scattering energies
        nene = (nesc + nprocs - 1) / nprocs     ! number of scattering energies managed by this image
        maxl = maxval(moldat % l2p)             ! highest partial wave angular momentum
        ntarg = size(moldat % etarg)            ! number of targes (residual ions)

        ! if not all final targets are required, reduce some of the above limits
        if (maxtarg > 0) then
            ntarg = min(ntarg, maxtarg)
            maxl = maxval(moldat % l2p, moldat % ichl <= maxtarg)
        end if

        ! find out how many absorption chains we have
        nchain = 0
        ptr => state
        do while (associated(ptr))
            if (ptr % order == order) then
                nchain = nchain + 1
            end if
            ptr => ptr % next
        end do

        allocate (omega(order), chains(order, nchain), cs(ntarg, nene), beta(2 + ntarg, nesc), &
                  M_xyz(nene, (maxl + 1)**2, nchain, ntarg), &
                  M_sph(nene, (maxl + 1)**2, nchain, ntarg))

#ifdef WITH_COARRAYS
        allocate (T((maxl + 1)**2, (maxl + 1)**2, nchain, nchain)[*])
#else
        allocate (T((maxl + 1)**2, (maxl + 1)**2, nchain, nchain))
#endif

        M_xyz = 0   ! multiphoton transition matrix elements in Cartesian basis
        M_sph = 0   ! multiphoton transition matrix elements in spherical basis

        ! the dipole elements stored in molecular_data do not contain the charge, so for each transition we need a factor of -1
        Q = (-1)**order

        ! assemble multi-photon molecular transition elements in Cartesian basis
        ichain = 0
        ptr => state
        do while (associated(ptr))
            if (ptr % order == order) then

                ! find this irreducible representation in molecular_data
                irr = findloc(moldat % mgvns, ptr % mgvn, 1)

                ! assemble polarisation components for this absorption chain
                ichain = ichain + 1
                parent => ptr
                do while (associated(parent % parent))
                    chains(parent % order, ichain) = mod(1 + parent % dcomp, 3) - 1  ! y ~ -1, z ~ 0, x ~ +1
                    parent => parent % parent
                end do

                ! copy all Cartesian matrix elements for this irreducible representation and absorption history to M_xyz
                do ichan = 1, size(ptr % dip, 1)
                    ii = moldat % ichl(ichan, irr)
                    li = moldat % l2p(ichan, irr)
                    mi = moldat % m2p(ichan, irr)
                    pw = li*li + li + mi + 1
                    mye = 0
                    do ie = myproc, nesc, nprocs
                        mye = mye + 1
                        Ek = escat(ie) - moldat % etarg(ii) + moldat % etarg(1)
                        if (Ek > 0) then
                            k = sqrt(2*Ek)
                            sigma = cphase(li, k)
                            d = cmplx(ptr % dip(ichan, 1, mye), ptr % dip(ichan, 2, mye), wp)
                            if (.not. d == d) d = 0  ! clear NaNs
                            M_xyz(mye, pw, ichain, ii) = imu**(-li) * (cos(sigma) + imu*sin(sigma)) * Q * d
                        end if
                    end do
                end do

            end if
            ptr => ptr % next
        end do

        if (raw == 'xyz' .or. raw == 'both') then
            call write_raw_dipoles(M_xyz, chains, nesc, 'xyz')
        end if

        ! transform the multiphoton transition matrix elements from Cartesian basis (M_xyz) to spherical basis (M_sph)
        do ii = 1, ntarg
            do l = 0, maxl
                do mi= -l, l
                    do mj = -l, l
                        pwi = l*l + l + mi + 1
                        pwj = l*l + l + mj + 1
                        do ichain = 1, nchain
                            do jchain = 1, nchain
                                cpl = conjg(sph_basis_transform_elm(l, mj, mi, 'Slm'))
                                do i = 1, order
                                    cpl = cpl * sph_basis_transform_elm(1, chains(i, jchain), chains(i, ichain), 'Slm')
                                end do
                                M_sph(:, pwj, jchain, ii) = M_sph(:, pwj, jchain, ii) + cpl * M_xyz(:, pwi, ichain, ii)
                            end do
                        end do
                    end do
                end do
            end do
        end do

        if (raw == 'sph' .or. raw == 'both') then
            call write_raw_dipoles(M_sph, chains, nesc, 'sph')
        end if

        ! hard-code laboratory-frame polarisations to zero
        p = 0

        ! evaluate and write the asymmetry parameters for all possible orders
        do L = 0, 2*order

            print '(/,A,I0)', 'Evaluating asymmetry parameter for L = ', L

            ! precalculate the contraction tensor
            i = 0
            do li = 0, maxl
                do mi = -li, li
                    pwi = li*li + li + mi + 1
                    do lj = 0, maxl
                        do mj = -lj, lj
                            pwj = lj*lj + lj + mj + 1
                            do ichain = 1, nchain
                                qi = chains(:, ichain)
                                do jchain = 1, nchain
                                    qj = chains(:, jchain)
                                    if (mod(i, nprocs) + 1 == myproc) then
                                        T(pwi, pwj, ichain, jchain) = beta_contraction_tensor(L, order, p, li, mi, qi, lj, mj, qj)
                                    else
                                        T(pwi, pwj, ichain, jchain) = 0
                                    end if
                                    i = i + 1
                                end do
                            end do
                        end do
                    end do
                end do
            end do

#ifdef WITH_COARRAYS
            ! allreduce the contraction tensor
            sync all
            call co_sum(T)
            sync all
#endif

            ! calculate partial cross sections and asymmetry parameters
            beta = 0
            do li = 0, maxl
                do mi = -li, li
                    pwi = li*li + li + mi + 1
                    do lj = 0, maxl
                        do mj = -lj, lj
                            pwj = lj*lj + lj + mj + 1
                            do ichain = 1, nchain
                                qi = chains(:, ichain)
                                do jchain = 1, nchain
                                    qj = chains(:, jchain)
                                    mye = 0
                                    do ie = myproc, nesc, nprocs
                                        mye = mye + 1
                                        do ii = 1, ntarg
                                            beta(2 + ii, mye) = beta(2 + ii, mye) + T(pwi, pwj, ichain, jchain) *  &
                                                real(M_sph(mye, pwi, ichain, ii) * conjg(M_sph(mye, pwj, jchain, ii)), wp)
                                        end do
                                    end do
                                end do
                            end do
                        end do
                    end do
                end do
            end do

            ! sum partial cross section, add prefactors
            mye = 0
            do ie = myproc, nesc, nprocs
                mye = mye + 1
                IP = escat(ie) + moldat % etarg(1) - Ei
                call calculate_photon_energies(IP, Ephoton, omega)
                beta(1, mye) = omega(1) * to_eV
                beta(2:, mye) = beta(2:, mye) * 2*pi * (2*pi*alpha)**order * product(omega)
                if (L == 0) then
                    cs(:, mye) = beta(3:, mye)       ! partial cross sections per target
                    beta(2, mye) = sum(cs(:, mye))   ! total cross section
                else
                    where (cs(:, mye) > 0)
                        beta(3:, mye) = beta(3:, mye) / cs(:, mye)  ! scale partial asymmetry parameter by the partial cross section
                    end where
                    beta(2, :) = 0                                ! there is no "total beta" for L > 0
                end if
            end do

            ! write the asymmetry parameter
            write (filename, '(A,I0,A,I0,A)') 'gen_', order, 'photo_beta_', L, '.txt'
            call write_cross_section(beta, nesc, filename)

        end do

    end subroutine calculate_asymmetry_parameters

#ifndef USE_BUILTIN_FINDLOC
    !> \brief   Find element in 4-byte integer array
    !> \author  J Benda
    !> \date    2021
    !>
    !> Compatibility replacement for the Fortran 2008 `findloc` function that is missing in GCC < 9.
    !> Only implements the one-dimensional variant for a specific combination of types.
    !>
    integer function findloc_32bit_array (array, element, dimen) result (idx)

        integer(int32), intent(in) :: array(:)
        integer,        intent(in) :: element, dimen

        idx = minloc(abs(array - element), 1)

    end function findloc_32bit_array

    !> \brief   Find element in 8-byte integer array
    !> \author  J Benda
    !> \date    2021
    !>
    !> Compatibility replacement for the Fortran 2008 `findloc` function that is missing in GCC < 9.
    !> Only implements the one-dimensional variant for a specific combination of types.
    !>
    integer function findloc_64bit_array (array, element, dimen) result (idx)

        integer(int64), intent(in) :: array(:)
        integer,        intent(in) :: element, dimen

        idx = minloc(abs(array - element), 1)

    end function findloc_64bit_array
#endif

end module multidip_routines
