! Copyright 2020
!
! For a comprehensive list of the developers that contributed to these codes
! see the UK-AMOR website.
!
! This file is part of UKRmol-out (UKRmol+ suite).
!
!     UKRmol-out is free software: you can redistribute it and/or modify
!     it under the terms of the GNU General Public License as published by
!     the Free Software Foundation, either version 3 of the License, or
!     (at your option) any later version.
!
!     UKRmol-out is distributed in the hope that it will be useful,
!     but WITHOUT ANY WARRANTY; without even the implied warranty of
!     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!     GNU General Public License for more details.
!
!     You should have received a copy of the GNU General Public License
!     along with  UKRmol-out (in source/COPYING). Alternatively, you can also visit
!     <https://www.gnu.org/licenses/>.
!
!> \brief   Special functions and objects used by MULTIDIP
!> \author  J Benda
!> \date    2020
!>
module multidip_special

    use, intrinsic :: iso_c_binding,   only: c_double, c_int, c_f_pointer
    use, intrinsic :: iso_fortran_env, only: input_unit, int32, int64, real64, output_unit

    use blas_lapack_gbl,  only: blasint
    use phys_const_gbl,   only: pi, imu
    use precisn_gbl,      only: wp

    implicit none

#ifdef WITH_GSL
    !> Derived type used by GSL for returning results
    type, bind(C) :: gsl_sf_result
        real(c_double) :: val
        real(c_double) :: err
    end type gsl_sf_result

    interface
        ! logarithm of the complex gamma function from GSL
        function gsl_sf_lngamma_complex_e (zr, zi, lnr, arg) bind(C, name='gsl_sf_lngamma_complex_e')
            use, intrinsic :: iso_c_binding, only: c_int, c_double
            import gsl_sf_result
            real(c_double), value :: zi, zr
            type(gsl_sf_result)   :: lnr, arg
            integer(c_int)        :: gsl_sf_lngamma_complex_e
        end function gsl_sf_lngamma_complex_e
    end interface

    interface
        ! Coulomb wave function from GSL
        function gsl_sf_coulomb_wave_FG_e (eta, rho, l, k, F, Fp, G, Gp, expF, expG) bind(C, name='gsl_sf_coulomb_wave_FG_e')
            use, intrinsic :: iso_c_binding, only: c_int, c_double
            import gsl_sf_result
            real(c_double), value :: eta, rho, l
            integer(c_int), value :: k
            type(gsl_sf_result)   :: F, Fp, G, Gp
            real(c_double)        :: expF, expG
            integer(c_int)        :: gsl_sf_coulomb_wave_FG_e
        end function gsl_sf_coulomb_wave_FG_e
    end interface

    interface
        ! irregular confluent hypergeometric function U
        pure function gsl_sf_hyperg_U (a, b, x) bind(C, name='gsl_sf_hyperg_U')
            use, intrinsic :: iso_c_binding, only: c_double
            real(c_double), value :: a, b, x
            real(c_double)        :: gsl_sf_hyperg_U
        end function gsl_sf_hyperg_U
    end interface
#endif

    interface
        ! Coulomb phase from UKRmol-out
        function cphaz (l, eta, iwrite)
            import wp
            integer  :: l, iwrite
            real(wp) :: eta, cphaz
        end function cphaz
    end interface

    interface
        ! Coulomb wave function from UKRmol-out
        subroutine coulfg (xx, eta1, xlmin, xlmax, fc, gc, fcp, gcp, mode1, kfn, ifail)
            import wp
            integer  :: mode1, kfn, ifail
            real(wp) :: xx, eta1, xlmin, xlmax, fc(*), gc(*), fcp(*), gcp(*)
        end subroutine coulfg
    end interface

    interface
        ! Decaying Whittaker function
        subroutine couln (l, Z, ERy, r, U, Up, acc, efx, ierr, nlast, fnorm)
            import wp
            integer  :: l, ierr, nlast
            real(wp) :: Z, Ery, r, U, Up, acc, efx, fnorm
        end subroutine couln
    end interface

    interface
        !Exponentially decaying solution with given (negative) energy and angular momentum
        subroutine decay (k, l, r, U, Up, iwrite)
            import wp
            integer  :: l, iwrite
            real(wp) :: k, r, U, Up
        end subroutine decay
    end interface

    interface
        ! real matrix-matrix multiplication from BLAS
        subroutine dgemm (transa, transb, m, n, k, alpha, A, lda, B, ldb, beta, C, ldc)
            import real64, blasint
            character(len=1) :: transa, transb
            integer(blasint) :: m, n, k, lda, ldb, ldc
            real(real64)     :: A(lda, *), B(ldb, *), C(ldc, *), alpha, beta
        end subroutine dgemm
    end interface

    interface
        ! real matrix-vector multiplication from BLAS
        subroutine dgemv (trans, m, n, alpha, A, lda, X, incx, beta, Y, incy)
            import real64, blasint
            character(len=1) :: trans
            integer(blasint) :: m, n, incx, incy, lda
            real(real64)     :: A(lda, *), X(*), Y(*), alpha, beta
        end subroutine dgemv
    end interface

    interface
        ! real LU decomposition from LAPACK
        subroutine dgetrf (m, n, A, lda, ipiv, info)
            import real64, blasint
            integer(blasint) :: m, n, lda, ipiv(*), info
            real(real64)     :: A(lda, *)
        end subroutine dgetrf
    end interface

    interface
        ! real LU backsubstitution from LAPACK
        subroutine dgetrs (trans, n, nrhs, A, lda, ipiv, B, ldb, info)
            import real64, blasint
            character(len=1) :: trans
            integer(blasint) :: n, nrhs, lda, ldb, info, ipiv(*)
            real(real64)     :: A(lda, *), B(ldb, *)
        end subroutine dgetrs
    end interface

    interface
        ! complex matrix-matrix multiplication from BLAS
        subroutine zgemm (transa, transb, m, n, k, alpha, A, lda, B, ldb, beta, C, ldc)
            import real64, blasint
            character(len=1) :: transa, transb
            integer(blasint) :: m, n, k, lda, ldb, ldc
            complex(real64)  :: A(lda, *), B(ldb, *), C(ldc, *), alpha, beta
        end subroutine zgemm
    end interface

    interface
        ! complex matrix-vector multiplication from BLAS
        subroutine zgemv (trans, m, n, alpha, A, lda, X, incx, beta, Y, incy)
            import real64, blasint
            character(len=1) :: trans
            integer(blasint) :: m, n, incx, incy, lda
            complex(real64)  :: A(lda, *), X(*), Y(*), alpha, beta
        end subroutine zgemv
    end interface

    interface
        ! complex LU decomposition from LAPACK
        subroutine zgetrf (m, n, A, lda, ipiv, info)
            import real64, blasint
            integer(blasint) :: m, n, lda, ipiv(*), info
            complex(real64)  :: A(lda, *)
        end subroutine zgetrf
    end interface

    interface
        ! complex LU backsubstitution from LAPACK
        subroutine zgetrs (trans, n, nrhs, A, lda, ipiv, B, ldb, info)
            import real64, blasint
            character(len=1) :: trans
            integer(blasint) :: n, nrhs, lda, ldb, info, ipiv(*)
            complex(real64)  :: A(lda, *), B(ldb, *)
        end subroutine zgetrs
    end interface

    interface
        ! complex matrix inversion from LAPACK
        subroutine zgetri (n, A, lda, ipiv, work, lwork, info)
            import real64, blasint
            integer(blasint) :: n, lda, ipiv(*), lwork, info
            complex(real64)  :: A(lda, *), work(*)
        end subroutine zgetri
    end interface

contains

    !> \brief   Calculate S-matrix from K-matrix
    !> \author  J Benda
    !> \date    2020
    !>
    !> Obtain the S-matrix from the definition formula
    !> ```
    !>    S = (1 + iK) (1 - iK)⁻¹
    !> ```
    !> Only the open-open sub-block of the K-matrix is used.
    !>
    subroutine calculate_S_matrix (Km, Sm, nopen)

        use multidip_params, only: cone, czero

        complex(wp), allocatable, intent(inout) :: Sm(:, :)
        real(wp),    allocatable, intent(in)    :: Km(:, :)
        integer,                  intent(in)    :: nopen

        complex(wp), allocatable :: IpiK(:, :), ImiK(:, :)
        integer(blasint)         :: i, m, lds

        m = nopen
        lds = size(Sm, 1)
        Sm(1:nopen, 1:nopen) = 0

        allocate (IpiK(nopen, nopen), ImiK(nopen, nopen))

        IpiK = 0
        ImiK = 0

        do i = 1, nopen
            IpiK(i, i) = 1
            ImiK(i, i) = 1
        end do

        IpiK = IpiK + imu*Km(1:nopen, 1:nopen)
        ImiK = ImiK - imu*Km(1:nopen, 1:nopen)

        call invert_matrix(ImiK)
        call zgemm('N', 'N', m, m, m, cone, IpiK, m, ImiK, m, czero, Sm, lds)  ! Sm(1:nopen, 1:nopen) = matmul(IpiK, ImiK)

    end subroutine calculate_S_matrix


    !> \brief   Photoionization coefficient
    !> \author  J Benda
    !> \date    2020
    !>
    !> Evaluates the wave function coefficient Ak for the final stationary photoionization wave.
    !>
    !> \param[in] rmatr  R-matrix radius for Coulomb wave matching
    !> \param[in] eig    Inner eigenenergies (R-matrix poles) for this irreducible representation
    !> \param[in] etot   Total energy of the system
    !> \param[in] w      Boundary amplitudes
    !> \param[in] Ek     Photoelectron kinetic energies in *all* channels (open and closed)
    !> \param[in] l      Photoelectron angular momentum in *all* channels (open and closed)
    !> \param[in] Km     K-matrix
    !> \param[inout] Ak  Wave function coefficients, needs to be allocated to neig x nopen and only that part will be written
    !>
    subroutine scatAk (rmatr, eig, etot, w, Ek, l, Km, Ak)

        use multidip_params,  only: rzero, rone

        complex(wp), intent(inout) :: Ak(:,:)
        real(wp),    intent(in)    :: rmatr, eig(:), etot, Ek(:), w(:,:), Km(:,:)
        integer,     intent(in)    :: l(:)

        complex(wp), allocatable :: T(:,:)
        real(wp),    allocatable :: V(:,:), TT(:,:,:), Sp(:,:), Cp(:,:), P(:), XX(:,:,:), wFp(:,:,:), k(:)

        integer(blasint) :: i, j, nchan, nstat, nopen, nochf, nochf2, ldwmp
        real(wp)         :: F, Fp, G, Gp

        nchan = size(Km, 1)         ! number of channels
        nopen = count(Ek > 0)       ! number of open channels
        nochf = size(Ak, 2)         ! number of channels for which to calcualte the Ak coefficients
        nstat = size(eig)           ! number of inner eigenstates for this irreducible representation
        ldwmp = size(w, 1)          ! leading dimension of the matrix where the boundary amplitudes are stored
        k = sqrt(2*abs(Ek))         ! linear momentum of the photoelectron

        nopen = min(nopen, nchan)
        nochf = min(nopen, nochf)
        nochf2 = 2*nochf

        allocate (V(nchan, nchan), T(nopen, nopen), Sp(nchan, nchan), Cp(nchan, nchan), TT(nchan, 2, nopen), &
                  XX(nchan, 2, nochf), wFp(nstat, 2, nochf), P(nstat))

        T = 0; V = 0; Sp = 0; Cp = 0

        ! construct, factorize and invert for T = (1 + iK)⁻¹ (nopen-by-nopen)
        do i = 1, nopen
            T(i,i) = 1
        end do
        T = T + imu*Km(1:nopen, 1:nopen)
        call invert_matrix(T)
        TT(1:nopen, 1, 1:nochf) = real(T(1:nopen, 1:nochf), wp)
        TT(1:nopen, 2, 1:nochf) = aimag(T(1:nopen, 1:nochf))

        ! evaluate the Coulomb functions and the normalization factor (nchan-by-nchan)
        do i = 1, nchan
            call coul(l(i), Ek(i), rmatr, F, Fp, G, Gp)
            V(i,i) = sqrt(2/(pi*k(i)))
            Sp(i,i) = Fp
            Cp(i,i) = Gp
        end do

        ! wave function coefficient (nstat-by-nochf)
        call dgemm('N', 'N', nchan, nopen,  nchan, rone, Cp, nchan, Km, nchan, rone,  Sp,  nchan)  ! S' + C' K
        call dgemm('N', 'N', nchan, nochf2, nopen, rone, Sp, nchan, TT, nchan, rzero, XX,  nchan)  ! (S' + C' K) T
        call dgemm('N', 'N', nchan, nochf2, nchan, rone, V,  nchan, XX, nchan, rzero, TT,  nchan)  ! V (S' + C' K) T
        call dgemm('T', 'N', nstat, nochf2, nchan, rone, w,  ldwmp, TT, nchan, rzero, wFp, nstat)  ! wT V (S' + C' K) T
        P = 0.5 / (eig - etot)
        do i = 1, nochf
            Ak(:, i) = P * cmplx(wFp(:, 1, i), wFp(:, 2, i), wp)
        end do

    end subroutine scatAk


    !> \brief   Coulomb phase
    !> \author  J Benda
    !> \date    2020
    !>
    !> Return the Coulomb phase, arg Gamma(l + 1 - i/k). Uses GSL or the UKRmol-out
    !> library, depending on the configuration.
    !>
    real(wp) function cphase (l, k) result (sigma)

        integer,  intent(in) :: l
        real(wp), intent(in) :: k
#ifdef WITH_GSL
        type(gsl_sf_result) :: lnr, arg
        integer(c_int)      :: err
        real(c_double)      :: zr, zi

        zr = l + 1
        zi = -1/k
        err = gsl_sf_lngamma_complex_e(zr, zi, lnr, arg)
        sigma = arg % val
#else
        sigma = cphaz(l, -1/k, 0)
#endif

    end function cphase


    !> \brief   Coulomb functions
    !> \author  J Benda
    !> \date    2020
    !>
    !> Evaluate the Coulomb wave (regular, irregular and derivatives). Uses GSL or the UKRmol-out
    !> library, depending on the configuration. For negative energies, evaluates the exponentially
    !> decreasing solution (into G and Gp) obtained from the Whittaker function (if charged) or
    !> from solution of the appropriate equation (if neutral).
    !>
    !> The derivatives returned are with respect to 'r' already, so they should not be multiplied
    !> by the additional factor of 'k'.
    !>
    subroutine coul (l, Ek, r, F, Fp, G, Gp)

        integer,  intent(in)    :: l
        real(wp), intent(in)    :: Ek, r
        real(wp), intent(inout) :: F, Fp, G, Gp
        real(wp) :: Z = 1
#ifdef WITH_GSL
        call coul_gsl(l, Ek, r, F, Fp, G, Gp)
#else
        call coul_ukrmol(l, Ek, r, F, Fp, G, Gp)
#endif
    end subroutine coul


    !> \brief   Coulomb-Hankel function
    !> \author  J Benda
    !> \date    2021
    !>
    !> See \ref coul for explanation of the arguments.
    !>
    complex(wp) function coulH (s, l, Ek, r) result(H)

        integer,  intent(in)    :: s, l
        real(wp), intent(in)    :: Ek, r

        real(wp) :: F, Fp, G, Gp

        call coul(l, Ek, r, F, Fp, G, Gp)

        H = cmplx(G, s*F, wp)

    end function coulH


    !> \brief   Coulomb-Hankel function (asymptotic form)
    !> \author  J Benda
    !> \date    2021
    !>
    !> Implements the asymptotic formula for Coulomb-Hankel function, DLMF §33.11.1.
    !> The number of terms is centrally controlled by a parameter in module multidip_params
    !> to be consistent with what is used in the integration routines.
    !>
    !> See \ref coul for explanation of the arguments.
    !>
    complex(wp) function coulH_asy (s, l, Ek, r) result(H)

        use multidip_params, only: nTermsAsy

        integer,  intent(in)    :: s, l
        real(wp), intent(in)    :: Ek, r

        integer     :: n
        complex(wp) :: a, b, term
        real(wp)    :: Z = 1, k

        k = sqrt(2*abs(Ek))

        if (Ek > 0) then
            a = l + 1 - Z*imu/k
            b = -l - Z*imu/k
        else
            a = l + 1 - Z/k
            b = -l - Z/k
        end if

        term = 1
        H = 1

        do n = 1, nTermsAsy
            if (Ek > 0) then
                term = term * (a + n - 1) * (b + n - 1) / (2*s*imu*n*k*r)
            else
                term = term * (a + n - 1) * (b + n - 1) / (-2*s*n*k*r)
            end if
            H = H + term
        end do

        if (Ek > 0) then
            H = H * exp(imu*s*(k*r + log(2*k*r)/k - pi*l/2 + cphase(l, k)))
        else
            H = H * exp(s*(-k*r + log(2*k*r)/k))
        end if

    end function coulH_asy


    !> \brief   Coulomb functions (GSL)
    !> \author  J Benda
    !> \date    2020
    !>
    !> Coulomb functions (positive- and negative-energy) calculated using GSL.
    !>
    subroutine coul_gsl (l, Ek, r, F, Fp, G, Gp)

        integer,  intent(in)    :: l
        real(wp), intent(in)    :: Ek, r
        real(wp), intent(inout) :: F, Fp, G, Gp
#ifdef WITH_GSL
        real(wp) :: Z = 1
        type(gsl_sf_result) :: resF, resFp, resG, resGp
        real(c_double)      :: expF, expG, eta, rho, lc, k, a, b, x, U, Up
        integer(c_int)      :: err, zero = 0

        if (Ek > 0) then
            ! positive-energy solution
            k = sqrt(2*Ek)
            rho = k*r
            eta = -Z/k
            lc = l
            err = gsl_sf_coulomb_wave_FG_e(eta, rho, lc, zero, resF, resFp, resG, resGp, expF, expG)
            F  = resF  % val
            Fp = resFp % val * k
            G  = resG  % val
            Gp = resGp % val * k
        else
            ! negative-energy solution, Whittaker function W (WARNING: charged only)
            k = sqrt(-2*Ek)
            a = l + 1 - Z/k
            b = 2*l + 2
            x = 2*k*r
            U = gsl_sf_hyperg_U(a, b, x)
            Up = -2*k*a * gsl_sf_hyperg_U(a + 1, b + 1, x)  ! DLMF §13.3.22
            F = 0
            Fp = 0
            G = exp(-x/2) * x**(l + 1) * U  ! DLMF §13.14.3
            Gp = exp(-x/2) * x**(l + 1) * (-k * U + (l + 1)/r * U + Up)
        end if
#else
        F = 0; Fp = 0
        G = 0; Gp = 0
#endif
    end subroutine coul_gsl


    !> \brief   Coulomb functions (UKRmol)
    !> \author  J Benda
    !> \date    2020
    !>
    !> Coulomb functions (positive- and negative-energy) calculated using UKRmol-out.
    !>
    subroutine coul_ukrmol (l, Ek, r, F, Fp, G, Gp)

        integer,  intent(in)    :: l
        real(wp), intent(in)    :: Ek, r
        real(wp), intent(inout) :: F, Fp, G, Gp
        real(wp) :: Z = 1
        integer  :: ifail, mode = 1, kfn = 0, nterms, iwrite = 0
        real(wp) :: fc(l + 1), gc(l + 1), fcp(l + 1), gcp(l + 1), lc, efx, fnorm, acc = 1e-10_wp, k

        lc = l
        if (Ek > 0) then
            ! positive-energy solution
            k = sqrt(2*Ek)
            call coulfg(k*r, -Z/k, lc, lc, fc, gc, fcp, gcp, mode, kfn, ifail)
            F = fc(l + 1)
            G = gc(l + 1)
            Fp = fcp(l + 1) * k
            Gp = gcp(l + 1) * k
        else
            ! negative energy solution (charged or neutral)
            if (Z /= 0) then
                call couln(l, Z, 2*Ek, r, G, Gp, acc, efx, ifail, nterms, fnorm)
            else
                call decay(2*Ek, l, r, G, Gp, iwrite)
            end if
            F = 0
            Fp = 0
        end if

    end subroutine coul_ukrmol


    !> \brief   Invert a complex matrix
    !> \author  J Benda
    !> \date    2020
    !>
    !> Calculate inverse of a complex matrix. Use the standard LAPACK decompose+solve sequence.
    !>
    subroutine invert_matrix (T)

        complex(wp), allocatable, intent(inout) :: T(:, :)

        complex(wp), allocatable :: work(:)
        complex(wp)              :: nwork(1)

        integer(blasint), allocatable :: ipiv(:)
        integer(blasint)              :: m, info, lwork

        m = size(T, 1)

        ! calculate LU decomposition
        allocate (ipiv(m))
        call zgetrf(m, m, T, m, ipiv, info)

        ! query for workspace size and allocate memory
        lwork = -1
        call zgetri(m, T, m, ipiv, nwork, lwork, info)
        lwork = int(nwork(1))
        allocate (work(lwork))

        ! compute the inverse
        call zgetri(m, T, m, ipiv, work, lwork, info)

    end subroutine invert_matrix


    !> \brief   Solve system of equations with complex matrix
    !> \author  J Benda
    !> \date    2020
    !>
    !> The matrix A is complex. The columns of matrices X and Y correspond to real and imaginary
    !> part of the right-hand side and of the solution, respectively.
    !>
    subroutine solve_complex_system (n, A, X, Y)

        complex(wp), allocatable, intent(inout) :: A(:, :)
        real(wp),    allocatable, intent(inout) :: X(:, :), Y(:, :)
        integer,                  intent(in)    :: n

        integer(blasint)              :: m, info, nrhs, lda
        integer(blasint), allocatable :: ipiv(:)
        complex(wp),      allocatable :: XX(:)

        allocate (ipiv(n))

        ! calculate LU decomposition
        m = n
        lda = size(A, 1)
        call zgetrf(m, m, A, lda, ipiv, info)

        ! combine real and imaginary part of the right-hand side
        XX = cmplx(X(1:n, 1), X(1:n, 2), wp)

        ! solve the equation
        nrhs = 1
        call zgetrs('N', m, nrhs, A, lda, ipiv, XX, m, info)

        Y(1:n, 1) = real(XX, wp)
        Y(1:n, 2) = aimag(XX)

    end subroutine solve_complex_system


    !> \brief   Construct or advance permutation
    !> \author  J Benda
    !> \date    2020
    !>
    !> If the given integer array contains a negative element, fill it with identical permutation (i.e.
    !> the sequence 1, 2, ..., N) and return TRUE. Otherwise attempt to generate the "next" permutation
    !> of the N values in the manner compatible with C++ `std::next_permutation` (lexicographically ordered
    !> sequences). When no further permutation is possible, return FALSE.
    !>
    !> The algorithm is shamelessly copied from the source of GCC's libstdc++ library and (except for the
    !> added initialization option) faithfully mimics the behaviour of the built-on C++ function `std::next_permutation`.
    !>
    logical function next_permutation (p) result (ok)

        integer, intent(inout) :: p(:)
        integer                :: i, j, k, n

        n = size(p)

        ! no permutation in empty array
        if (n == 0) then
            ok = .false.
            return
        end if

        ! initialize a new permutation
        if (any(p < 1)) then
            do i = 1, n
                p(i) = i
            end do
            ok = .true.
            return
        end if

        ! one-element permutation cannot be advanced further
        if (n == 1) then
            ok = .false.
            return
        end if

        i = n

        do
            k = i
            i = i - 1
            if (p(i) < p(k)) then
                j = n
                do while (p(i) >= p(j))
                    j = j - 1
                end do
                call swap(p(i), p(j))
                call reverse(p(k:n))
                ok = .true.
                return
            else if (i == 1) then
                ok = .false.
                return
            end if
        end do

    end function next_permutation


    !> \brief   Exchange value of two integers
    !> \author  J Benda
    !> \date    2020
    !>
    !> This subroutine mimics the behaviour of the built-in C++ function std::swap.
    !>
    subroutine swap (a, b)

        integer, intent(inout) :: a, b
        integer                :: c

        c = a
        a = b
        b = c

    end subroutine swap


    !> \brief   Reverse order of elements in array
    !> \author  J Benda
    !> \date    2020
    !>
    !> This subroutine mimics the behaviour of the built-in C++ function `std::reverse`.
    !>
    subroutine reverse (a)

        integer, intent(inout) :: a(:)
        integer                :: i, n

        n = size(a)

        do i = 1, n/2
            call swap(a(i), a(n + 1 - i))
        end do

    end subroutine reverse


    !> \brief   Compensated summation
    !> \author  J Benda
    !> \date    2020
    !>
    !> Add dX to X, keep track of numerical error. Uses Kahan's algorithm. This subroutine is in infinite
    !> precision equivalent to just
    !> ```
    !>     X = X + dX
    !>     err = 0
    !> ```
    !> but in the finite precision arithmetic it compensates the running numerical error. It mustn't be
    !> optimized away by the compiler! Flags like `-ffast-math` or `-Ofast` are detrimental here. Common
    !> optimization flags like `-O2` or `-O3` should be safe (but it may depend on the compiler).
    !>
    subroutine kahan_add (X, dX, err)

        complex(wp), intent(inout) :: X, err
        complex(wp), intent(in)    :: dX

        complex(wp) :: Y, Z

        Y = dX - err
        Z = X + Y
        err = (Z - X) - Y  ! the parenthesis must be obeyed, otherwise we get rubbish
        X = Z

    end subroutine kahan_add


    !> \brief   Angular part of the beta parameters
    !> \author  J Benda
    !> \date    2020
    !>
    !> Evaluates the contraction coefficient \f$ T_{m p_1 q_1 q_1'\dots p_n q_n q_n'}^{Jll'} \f$ in
    !> \f[
    !>     \frac{\mathrm{d}\sigma^{(p_1\dots p_n)}}{\mathrm{d}\Omega} = \frac{1}{4\pi} \sum_{J} b_J^{(p_1\dots p_n)} P_J(\cos\theta)
    !>         = \frac{1}{4\pi} \sum_{Jlml'm'\atop q_1 q_1' \dots q_n q_n'}
    !>         M_{mq_1\dots q_n}^{(l)} M_{m'q_1'\dots q_n'}^{(l')*} T_{m m', p_1 q_1 q_1'\dots p_n q_n q_n'}^{Jll'} P_J(\cos\theta),
    !> \f]
    !> which has the following explicit form:
    !> \f[
    !>     T_{m m', p_1 q_1 q_1'\dots p_n q_n q_n'}^{Jll'} = \sum_{\mu} (-1)^{\mu} (2J + 1) \sqrt{(2l + 1)(2l' + 1)}
    !>         \left(\begin{matrix} l & l' & J \\ 0 & 0 & 0 \end{matrix}\right)
    !>         \left(\begin{matrix} l & l' & J \\ \mu & -\mu & 0 \end{matrix}\right)
    !>         \frac{1}{8\pi^2} \int D_{\mu m}^{(l)} D_{\mu m'}^{(l')*} D_{p_1 q_1}^{(1)*} D_{p_1 q_1'}^{(1)} \dots
    !>            D_{p_n q_n}^{(1)*} D_{p_n q_n}^{(1)} \mathrm{d}\hat{R} .
    !> \f]
    !> The real averaging integral over orientations of the molecule is re-expressed as
    !> \f[
    !>     (-1)^{m + m'}
    !>     \frac{1}{8\pi^2} \int D_{-\mu, -m}^{(l)*} D_{-\mu, -m'}^{(l')} D_{p_1 q_1}^{(1)*} D_{p_1 q_1'}^{(1)} \dots
    !>            D_{p_n q_n}^{(1)*} D_{p_n q_n}^{(1)} \mathrm{d}\hat{R}
    !> \f]
    !> and then calculated in \ref wigner_D_multiint.
    !>
    !> \note In the present implementation, all laboratory-frame polarisations \f$ p_i \f$ are considered equal to zero, i.e.
    !>       the field is linearly polarized.
    !>
    real(wp) function beta_contraction_tensor (J, n, p, li, mi, qi, lj, mj, qj) result (T)

        use dipelm_special_functions, only: threej
        use multidip_params,          only: rone

        integer, intent(in) :: J, n, p(n), li, mi, qi(n), lj, mj, qj(n)
        integer :: mu, l(n + 1), lp(n + 1), a(n + 1), ap(n + 1), b(n + 1), bp(n + 1)
        real(wp) :: I, Q

        l(1) = li; l(2:) = 1
        lp(1) = lj; lp(2:) = 1

        b(1) = -mi; b(2:) = qi
        bp(1) = -mj; bp(2:) = qj

        T = 0

        do mu = -li, li

            a(1) = -mu; a(2:) = p
            ap(1) = -mu; ap(2:) = p

            I = (-1)**(mi + mj) * wigner_D_multiint(n + 1, l, a, b, lp, ap, bp)
            Q = (-1)**mu * (2*J + 1) * sqrt((2*li + rone)*(2*lj + rone))  &
                    * threej(2*li, 0, 2*lj, 0, 2*J, 0) * threej(2*li, 2*mu, 2*lj, -2*mu, 2*J, 0)

            T = T + I*Q

        end do

    end function beta_contraction_tensor


    !> \brief   Orientation averaging
    !> \author  J Benda
    !> \date    2020
    !>
    !> Calculate the value of the integral
    !> \f[
    !>    \frac{1}{8\pi^2} \int D_{a_1 b_1}^{(l_1)} D_{a_1' b_1'}^{(l_1')*} \dots
    !>                          D_{a_1 b_1}^{(l_n)} D_{a_n' b_n'}^{(l_n')*} \mathrm{d}\hat{R}
    !> \f]
    !> over all orientations (Euler angles) \f$ \hat{R} \f$. The calculation is done recursively using the formula
    !> \f[
    !>    D_{a_1 b_1}^{(l_1)} D_{a_2 b_2}^{(l_2)} = \sum_{juv} (-1)^{u - v} (2j + 1)
    !>     \left(\begin{matrix} l_1 & l_2 & j \\ a_1 & a_2 & -u \end{matrix}\right)
    !>     \left(\begin{matrix} l_1 & l_2 & j \\ b_1 & b_2 & -v \end{matrix}\right)
    !>     D_{u v}^{(j)}
    !> \f]
    !> that is applied to \f$ D_{p_1 q_1}^{(l_1)} D_{p_2 q_2}^{(l_2)} \f$ and to \f$ D_{p_1 q_1'}^{(l_1)} D_{p_2 q_2'}^{(l_2)} \f$,
    !> leading to the linear combination of shorter integrals
    !> \f[
    !>    \frac{1}{8\pi^2} \int D_{u v}^{(j)} D_{u' v'}^{(j')*} D_{a_3 b_3}^{(l_3)} D_{a_3' b_3'}^{(l_3)*}
    !>                          \dots D_{a_n b_n}^{(l_n)} D_{a_n' b_n'}^{(l_n)*} \mathrm{d}\hat{R}
    !> \f]
    !> The recursion terminates when only the pair \f$ D_{u v}^{(j)} D_{u' v'}^{(j')*} \f$ is left,
    !> by means of the orthogonality relation
    !> \f[
    !>     \frac{1}{8\pi^2} \int D_{u v}^{(j)} D_{u' v'}^{(j')*} \mathrm{d}\hat{R}
    !>     = \frac{1}{2j + 1} \delta_{jj'} \delta_{uu'} \delta_{vv¨'} \,.
    !> \f]
    !>
    recursive real(wp) function wigner_D_multiint (n, l, a, b, lp, ap, bp) result (K)

        use dipelm_special_functions, only: threej
        use multidip_params,          only: rone

        integer, intent(in) :: n
        integer, intent(in) :: l(n), a(n), b(n), lp(n), ap(n), bp(n)

        integer  :: u(n - 1), v(n - 1), up(n - 1), vp(n - 1), j(n - 1), jp(n - 1), j1, jp1
        real(wp) :: wa, wb, wap, wbp, W

        K = 0

        ! orthogonality relation
        if (n == 1) then
            if (l(1) == lp(1) .and. a(1) == ap(1) .and. b(1) == bp(1)) K = 1 / (2*l(1) + rone)
            return
        end if

        u(1) = a(1) + a(2); u(2:) = a(3:)
        v(1) = b(1) + b(2); v(2:) = b(3:)

        up(1) = ap(1) + ap(2); up(2:) = ap(3:)
        vp(1) = bp(1) + bp(2); vp(2:) = bp(3:)

        j(2:) = l(3:)
        jp(2:) = lp(3:)

        ! recursion: couple the first two angular momenta (both primed and unprimed)
        do j1 = max(abs(u(1)), max(abs(v(1)), abs(l(1) - l(2)))), l(1) + l(2)
            do jp1 = max(abs(up(1)), max(abs(vp(1)), abs(lp(1) - lp(2)))), lp(1) + lp(2)

                j(1) = j1
                jp(1) = jp1

                wa  = threej(2*l(1), 2*a(1), 2*l(2), 2*a(2), 2*j(1), -2*u(1))
                wb  = threej(2*l(1), 2*b(1), 2*l(2), 2*b(2), 2*j(1), -2*v(1))
                wap = threej(2*lp(1), 2*ap(1), 2*lp(2), 2*ap(2), 2*jp(1), -2*up(1))
                wbp = threej(2*lp(1), 2*bp(1), 2*lp(2), 2*bp(2), 2*jp(1), -2*vp(1))
                W = wigner_D_multiint(n - 1, j, u, v, jp, up, vp)
                K = K + (-1)**(u(1) - v(1) + up(1) - vp(1)) * (2*j1 + 1) * (2*jp1 + 1) * wa * wap * wb * wbp * W

            end do
        end do

    end function wigner_D_multiint


    !> \brief   Two-photon asymmetry parameter
    !> \author  J Benda
    !> \date    2020
    !>
    !> Explicit form of the two-photon asymmetry parameter from Demekhin, Lagutin, Petrov, Phys. Rev. A 85 (2012) 023416.
    !>
    !> \note The explicit formula contains an apparent extra factor of (-1)^(L + lj), which is however fully absorbed
    !>       into the phase of the wave function in the present approach.
    !>
    real(wp) function beta_2p_demekhin (L, p1, p2, li, mi, q1i, q2i, lj, mj, q1j, q2j) result (b)

        use dipelm_special_functions, only: threej
        use multidip_params,          only: rone

        integer, intent(in) :: L, p1, p2, li, mi, lj, mj, q1i, q1j, q2i, q2j

        integer  :: M_L, J, M_J, K, M_K
        real(wp) :: w(6)

        b = 0
        M_L = -(mi - mj)

        do J = 0, 2
            do M_J = -J, +J
                do K = 0, 2
                    do M_K = -K, +K
                        w(1) = threej(2*1, 2*p1,  2*1, -2*p1,  2*J,    0)
                        w(2) = threej(2*1, 2*q1i, 2*1, -2*q1j, 2*J, -2*M_J)
                        w(3) = threej(2*1, 2*q2i, 2*1, -2*q2j, 2*K, -2*M_K)
                        w(4) = threej(2*1, 2*p2,  2*1, -2*p2,  2*K,    0)
                        w(5) = threej(2*J, 2*M_J, 2*K,  2*M_K, 2*L,  2*M_L)
                        w(6) = threej(2*J,   0,   2*K,    0,   2*L,    0)
                        b = b + (-1)**(q1j + q2j) * (2*J + 1) * (2*K + 1) * product(w(1:6))
                    end do
                end do
            end do
        end do

        w(1) = threej(2*li,    0,  2*lj,   0,  2*L,    0)
        w(2) = threej(2*li, -2*mi, 2*lj, 2*mj, 2*L, -2*M_L)

        b = b * (-1)**mi * (2*L + 1) * sqrt((2*li + rone)*(2*lj + rone)) * product(w(1:2))

    end function beta_2p_demekhin

end module multidip_special
